# Copyright (C) 2006-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Aggregate needed object instances for checker threads.
"""
import threading

import requests
import time
import urllib.parse
import random
from .. import log, LOG_CHECK, strformat, LinkCheckerError
from ..decorators import synchronized
from ..cache import urlqueue
from ..htmlutil import loginformsearch
from ..cookies import from_file
from . import logger, status, checker, interrupter


_threads_lock = threading.RLock()
_hosts_lock = threading.RLock()
_downloadedbytes_lock = threading.RLock()


def new_request_session(config, cookies):
    """Create a new request session."""
    session = requests.Session()
    if cookies:
        session.cookies = cookies
    session.max_redirects = config["maxhttpredirects"]
    session.headers.update(
        {"User-Agent": config["useragent"]}
    )
    if config["cookiefile"]:
        for cookie in from_file(config["cookiefile"]):
            session.cookies.set_cookie(cookie)
    return session


class Aggregate:
    """Store thread-safe data collections for checker threads."""
    wait_time_min_default = 0.1
    wait_time_max_default = 0.6

    def __init__(self, config, urlqueue, robots_txt, plugin_manager, result_cache):
        """Store given link checking objects."""
        self.config = config
        self.urlqueue = urlqueue
        self.logger = logger.Logger(config)
        self.threads = []
        self.request_sessions = {}
        self.robots_txt = robots_txt
        self.plugin_manager = plugin_manager
        self.result_cache = result_cache
        self.times = {}
        self.maxrated = {}
        self.cookies = None
        requests_per_second = config["maxrequestspersecond"]
        self.wait_time_min = 1.0 / requests_per_second
        self.wait_time_max = 6 * self.wait_time_min
        self.downloaded_bytes = 0

    def visit_loginurl(self):
        """Check for a login URL and visit it."""
        url = self.config["loginurl"]
        if not url:
            return
        user, password = self.config.get_user_password(url)
        if not user and not password:
            raise LinkCheckerError(
                "loginurl is configured but neither user nor password are set"
            )
        session = new_request_session(self.config, self.cookies)
        log.debug(LOG_CHECK, "Getting login form %s", url)
        kwargs = dict(timeout=self.config["timeout"])
        # XXX: sslverify?  can we reuse HttpUrl.get_request_kwargs()
        # somehow?
        response = session.get(url, **kwargs)
        response.raise_for_status()
        cgiuser = self.config["loginuserfield"] if user else None
        cgipassword = self.config["loginpasswordfield"] if password else None
        form = loginformsearch.search_form(response.text, cgiuser, cgipassword)
        if not form:
            raise LinkCheckerError("Login form not found at %s" % url)
        if user:
            form.data[cgiuser] = user
        if password:
            form.data[cgipassword] = password
        for key, value in self.config["loginextrafields"].items():
            form.data[key] = value
        formurl = urllib.parse.urljoin(url, form.url)
        log.debug(LOG_CHECK, "Posting login data to %s", formurl)
        response = session.post(formurl, data=form.data, **kwargs)
        response.raise_for_status()
        self.cookies = session.cookies
        if len(self.cookies) == 0:
            raise LinkCheckerError("No cookies set by login URL %s" % url)

    @synchronized(_threads_lock)
    def start_threads(self):
        """Spawn threads for URL checking and status printing."""
        if self.config["status"]:
            t = status.Status(self, self.config["status_wait_seconds"])
            t.start()
            self.threads.append(t)
        if self.config["maxrunseconds"]:
            t = interrupter.Interrupt(self.config["maxrunseconds"])
            t.start()
            self.threads.append(t)
        num = self.config["threads"]
        if num > 0:
            for dummy in range(num):
                t = checker.Checker(
                    self.urlqueue, self.logger, self.add_request_session
                )
                self.threads.append(t)
                t.start()
        else:
            self.request_sessions[threading.get_ident()] = new_request_session(
                self.config, self.cookies
            )
            checker.check_urls(self.urlqueue, self.logger)

    @synchronized(_threads_lock)
    def add_request_session(self):
        """Add a request session for current thread."""
        session = new_request_session(self.config, self.cookies)
        self.request_sessions[threading.get_ident()] = session

    @synchronized(_threads_lock)
    def get_request_session(self):
        """Get the request session for current thread."""
        return self.request_sessions[threading.get_ident()]

    @synchronized(_hosts_lock)
    def wait_for_host(self, host):
        """Throttle requests to one host."""
        t = time.time()
        if host in self.times:
            due_time = self.times[host]
            if due_time > t:
                wait = due_time - t
                time.sleep(wait)
                t = time.time()
        if host in self.maxrated:
            wait_time_min, wait_time_max = self.wait_time_min, self.wait_time_max
        else:
            wait_time_min = max(self.wait_time_min, self.wait_time_min_default)
            wait_time_max = max(self.wait_time_max, self.wait_time_max_default)
        log.debug(LOG_CHECK,
                  "Min wait time: %s Max wait time: %s for host: %s",
                  wait_time_min, wait_time_max, host)
        wait_time = random.uniform(wait_time_min, wait_time_max)
        self.times[host] = t + wait_time

    @synchronized(_hosts_lock)
    def set_maxrated_for_host(self, host):
        """Remove the limit on the maximum request rate for a host."""
        self.maxrated[host] = True

    @synchronized(_threads_lock)
    def print_active_threads(self):
        """Log all currently active threads."""
        debug = log.is_debug(LOG_CHECK)
        if debug:
            first = True
            for name in self.get_check_threads():
                if first:
                    log.info(LOG_CHECK, _("These URLs are still active:"))
                    first = False
                log.info(LOG_CHECK, name[12:])
        args = dict(
            num=len(
                [x for x in self.threads if x.name.startswith("CheckThread-")]
            ),
            timeout=strformat.strduration_long(self.config["aborttimeout"]),
        )
        log.info(
            LOG_CHECK,
            _(
                "%(num)d URLs are still active. After a timeout of %(timeout)s"
                " the active URLs will stop."
            )
            % args,
        )

    @synchronized(_threads_lock)
    def get_check_threads(self):
        """Return iterator of checker threads."""
        for t in self.threads:
            if t.name.startswith("CheckThread-"):
                yield t.name

    def cancel(self):
        """Empty the URL queue."""
        self.urlqueue.do_shutdown()

    def abort(self):
        """Print still-active URLs and empty the URL queue."""
        self.print_active_threads()
        self.cancel()
        timeout = self.config["aborttimeout"]
        try:
            self.urlqueue.join(timeout=timeout)
        except urlqueue.Timeout:
            log.warn(
                LOG_CHECK,
                "Abort timed out after %d seconds, stopping application." % timeout,
            )
            raise KeyboardInterrupt()

    @synchronized(_threads_lock)
    def remove_stopped_threads(self):
        """Remove the stopped threads from the internal thread list."""
        self.threads = [t for t in self.threads if t.is_alive()]

    @synchronized(_threads_lock)
    def finish(self):
        """Wait for checker threads to finish."""
        if not self.urlqueue.empty():
            # This happens when all checker threads died.
            self.cancel()
        for t in self.threads:
            t.stop()
        for t in self.threads:
            t.join(timeout=1.0)

    @synchronized(_threads_lock)
    def is_finished(self):
        """Determine if checking is finished."""
        self.remove_stopped_threads()
        return self.urlqueue.empty() and not self.threads

    @synchronized(_downloadedbytes_lock)
    def add_downloaded_bytes(self, numbytes):
        """Add to number of downloaded bytes."""
        self.downloaded_bytes += numbytes

    def end_log_output(self, **kwargs):
        """Print ending output to log."""
        kwargs.update(
            dict(
                downloaded_bytes=self.downloaded_bytes, num_urls=len(self.result_cache),
            )
        )
        self.logger.end_log_output(**kwargs)



# Copyright (C) 2006-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Cache robots.txt contents.
"""
import urllib.parse

from .. import robotparser2
from ..containers import LFUCache
from ..decorators import synchronized
from ..lock import get_lock
from .. import log, LOG_CACHE


# lock objects
cache_lock = get_lock("robots.txt_cache_lock")
robot_lock = get_lock("robots.txt_robot_lock")


class RobotsTxt:
    """
    Thread-safe cache of downloaded robots.txt files.
    format: {cache key (string) -> robots.txt content (RobotFileParser)}
    """

    def __init__(self, useragent):
        """Initialize per-URL robots.txt cache."""
        # mapping {URL -> parsed robots.txt}
        self.cache = LFUCache(size=100)
        self.hits = self.misses = 0
        self.roboturl_locks = {}
        self.useragent = useragent

    def allows_url(self, url_data, timeout=None):
        """Ask robots.txt allowance."""
        roboturl = url_data.get_robots_txt_url()
        with self.get_lock(roboturl):
            return self._allows_url(url_data, roboturl, timeout)

    def _allows_url(self, url_data, roboturl, timeout=None):
        """Ask robots.txt allowance. Assumes only single thread per robots.txt
        URL calls this function."""
        with cache_lock:
            if roboturl in self.cache:
                self.hits += 1
                rp = self.cache[roboturl]
                return rp.can_fetch(self.useragent, url_data.url)
            self.misses += 1
        kwargs = dict(auth=url_data.auth, session=url_data.session, timeout=timeout)
        rp = robotparser2.RobotFileParser(**kwargs)
        rp.set_url(roboturl)
        rp.read()
        with cache_lock:
            self.cache[roboturl] = rp
        self.add_sitemap_urls(rp, url_data, roboturl)
        return rp.can_fetch(self.useragent, url_data.url)

    def add_sitemap_urls(self, rp, url_data, roboturl):
        """Add sitemap URLs to queue."""
        if not rp.sitemap_urls or not url_data.allows_simple_recursion():
            return
        for sitemap_url, line in rp.sitemap_urls:
            if not urllib.parse.urlparse(sitemap_url).scheme:
                log.warn(LOG_CACHE, _("Relative Sitemap %s in %s discarded"),
                         sitemap_url, roboturl)
                continue
            url_data.add_url(sitemap_url, line=line, parent=roboturl)

    @synchronized(robot_lock)
    def get_lock(self, roboturl):
        """Return lock for robots.txt url."""
        return self.roboturl_locks.setdefault(roboturl, get_lock(roboturl))



# Copyright (C) 2000-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Base URL handler.
"""
import sys
import os
import urllib.parse
from urllib.request import urlopen
import time
import errno
import socket
from io import BytesIO

from . import absolute_url, get_url_from
from .. import (
    log,
    LOG_CHECK,
    strformat,
    LinkCheckerError,
    url as urlutil,
    trace,
    get_link_pat,
)
from ..htmlutil import htmlsoup
from ..network import iputil
from .const import (
    WARN_URL_EFFECTIVE_URL,
    WARN_URL_ERROR_GETTING_CONTENT,
    WARN_URL_OBFUSCATED_IP,
    WARN_URL_CONTENT_SIZE_ZERO,
    WARN_URL_CONTENT_SIZE_TOO_LARGE,
    WARN_URL_CONTENT_TYPE_UNPARSEABLE,
    WARN_URL_WHITESPACE,
    URL_MAX_LENGTH,
    WARN_URL_TOO_LONG,
    ExcList,
    ExcSyntaxList,
    ExcNoCacheList,
)
from ..url import url_fix_wayback_query

# schemes that are invalid with an empty hostname
scheme_requires_host = ("ftp", "http", "telnet")


def urljoin(parent, url):
    """
    If url is relative, join parent and url. Else leave url as-is.

    @return: joined url
    """
    if urlutil.url_is_absolute(url):
        return url
    return urllib.parse.urljoin(parent, url)


def url_norm(url, encoding):
    """Wrapper for url.url_norm() to convert UnicodeError in
    LinkCheckerError."""
    try:
        return urlutil.url_norm(url, encoding=encoding)
    except UnicodeError:
        msg = _("URL has unparsable domain name: %(name)s") % {
            "name": sys.exc_info()[1]
        }
        raise LinkCheckerError(msg)


class UrlBase:
    """An URL with additional information like validity etc."""

    # file types that can be parsed recursively
    ContentMimetypes = {
        "text/html": "html",
        "application/xhtml+xml": "html",
        # Include PHP file which helps when checking local .php files.
        # It does not harm other URL schemes like HTTP since HTTP servers
        # should not send this content type. They send text/html instead.
        "application/x-httpd-php": "html",
        "text/css": "css",
        "application/vnd.adobe.flash.movie": "swf",
        "application/x-shockwave-flash": "swf",
        "application/msword": "word",
        "text/plain+linkchecker": "text",
        "text/plain+opera": "opera",
        "text/plain+chromium": "chromium",
        "application/x-plist+safari": "safari",
        "text/vnd.wap.wml": "wml",
        "application/xml+sitemap": "sitemap",
        "application/xml+sitemapindex": "sitemapindex",
        "application/pdf": "pdf",
        "application/x-pdf": "pdf",
    }

    # Read in 16kb chunks
    ReadChunkBytes = 1024 * 16

    def __init__(
        self,
        base_url,
        recursion_level,
        aggregate,
        parent_url=None,
        base_ref=None,
        line=-1,
        column=-1,
        page=-1,
        name="",
        url_encoding=None,
        extern=None,
    ):
        """
        Initialize check data, and store given variables.

        @param base_url: unquoted and possibly unnormed url
        @param recursion_level: on what check level lies the base url
        @param aggregate: aggregate instance
        @param parent_url: quoted and normed url of parent or None
        @param base_ref: quoted and normed url of <base href=""> or None
        @param line: line number of url in parent content
        @param column: column number of url in parent content
        @param page: page number of url in parent content
        @param name: name of url or empty
        @param url_encoding: encoding of URL or None
        @param extern: None or (is_extern, is_strict)
        """
        self.reset()
        self.init(
            base_ref,
            base_url,
            parent_url,
            recursion_level,
            aggregate,
            line,
            column,
            page,
            name,
            url_encoding,
            extern,
        )
        self.check_syntax()
        if recursion_level == 0:
            self.add_intern_pattern()
        self.set_extern(self.url)
        if self.extern[0] and self.extern[1]:
            self.add_info(
                _("The URL is outside of the domain filter, checked only syntax.")
            )
            if not self.has_result:
                self.set_result(_("filtered"))

    def init(
        self,
        base_ref,
        base_url,
        parent_url,
        recursion_level,
        aggregate,
        line,
        column,
        page,
        name,
        url_encoding,
        extern,
    ):
        """
        Initialize internal data.
        """
        self.base_ref = base_ref
        if self.base_ref is not None:
            assert isinstance(self.base_ref, str), repr(self.base_ref)
        self.base_url = base_url.strip() if base_url else base_url
        if self.base_url is not None:
            assert isinstance(self.base_url, str), repr(self.base_url)
        self.parent_url = parent_url
        if self.parent_url is not None:
            assert isinstance(self.parent_url, str), repr(self.parent_url)
        self.recursion_level = recursion_level
        self.aggregate = aggregate
        self.line = line
        self.column = column
        self.page = page
        self.name = name
        assert isinstance(self.name, str), repr(self.name)
        self.encoding = url_encoding
        self.extern = extern
        if self.base_ref:
            assert not urlutil.url_needs_quoting(self.base_ref), (
                "unquoted base reference URL %r" % self.base_ref
            )
        if self.parent_url:
            assert not urlutil.url_needs_quoting(self.parent_url), (
                "unquoted parent URL %r" % self.parent_url
            )
        url = absolute_url(self.base_url, base_ref, parent_url)
        # assume file link if no scheme is found
        self.scheme = url.split(":", 1)[0].lower() or "file"
        if self.base_url != base_url:
            self.add_warning(
                _("Leading or trailing whitespace in URL `%(url)s'.")
                % {"url": base_url},
                tag=WARN_URL_WHITESPACE,
            )
        self.ignore_errors = self.aggregate.config['ignoreerrors']

    def reset(self):
        """
        Reset all variables to default values.
        """
        # self.url is constructed by self.build_url() out of base_url
        # and (base_ref or parent) as absolute and normed url.
        # This the real url we use when checking so it also referred to
        # as 'real url'
        self.url = None
        # a split version of url for convenience
        self.urlparts = None
        # the scheme, host, port and anchor part of url
        self.scheme = self.host = self.port = self.anchor = None
        # the result message string and flag
        self.result = ""
        self.has_result = False
        # valid or not
        self.valid = True
        # list of warnings (without duplicates)
        self.warnings = []
        # list of infos
        self.info = []
        # content size
        self.size = -1
        # last modification time of content in HTTP-date format
        # as specified in RFC2616 chapter 3.3.1
        self.modified = None
        # download time
        self.dltime = -1
        # check time
        self.checktime = 0
        # connection object
        self.url_connection = None
        # data of url content,  (data == None) means no data is available
        self.data = None
        # url content data encoding
        self.content_encoding = None
        # url content as a Unicode string
        self.text = None
        # url content as a Beautiful Soup object
        self.soup = None
        # cache url is set by build_url() calling set_cache_url()
        self.cache_url = None
        # extern flags (is_extern, is_strict)
        self.extern = None
        # flag if the result should be cached
        self.caching = True
        # title is either the URL or parsed from content
        self.title = None
        # flag if content should be checked or not
        self.do_check_content = True
        # MIME content type
        self.content_type = ""
        # URLs seen through redirections
        self.aliases = []
        # error messages (regular expressions) to ignore
        self.ignore_errors = []

    def set_result(self, msg, valid=True, overwrite=False):
        """
        Set result string and validity.
        """
        if self.has_result and not overwrite:
            log.warn(
                LOG_CHECK,
                "Double result %r (previous %r) for %s",
                msg,
                self.result,
                self,
            )
        else:
            self.has_result = True
        if not msg:
            log.warn(LOG_CHECK, "Empty result for %s", self)
        self.result = msg
        self.valid = valid

        if not self.valid:
            for url_regex, msg_regex in self.ignore_errors:
                if not url_regex.search(self.url):
                    continue
                if not msg_regex.search(self.result):
                    continue
                self.valid = True
                self.result = f"Ignored: {self.result}"

        # free content data
        self.data = None

    def get_title(self):
        """Return title of page the URL refers to.
        This is per default the filename or the URL."""
        if self.title is None:
            url = ""
            if self.base_url:
                url = self.base_url
            elif self.url:
                url = self.url
            self.title = url
            if "/" in url:
                title = url.rsplit("/", 1)[1]
                if title:
                    self.title = title
        return self.title

    def is_content_type_parseable(self):
        """
        Return True iff the content type of this url is parseable.
        """
        if self.content_type in self.ContentMimetypes:
            return True
        log.debug(
            LOG_CHECK,
            "URL with content type %r is not parseable",
            self.content_type,
        )
        if self.recursion_level == 0:
            self.add_warning(
                _("The URL with content type %r is not parseable.") % self.content_type,
                tag=WARN_URL_CONTENT_TYPE_UNPARSEABLE,
            )
        return False

    def is_parseable(self):
        """
        Return True iff content of this url is parseable.
        """
        return False

    def is_html(self):
        """Return True iff content of this url is HTML formatted."""
        return self._is_ctype("html")

    def is_css(self):
        """Return True iff content of this url is CSS stylesheet."""
        return self._is_ctype("css")

    def _is_ctype(self, ctype):
        """Return True iff content is valid and of the given type."""
        if not self.valid:
            return False
        mime = self.content_type
        return self.ContentMimetypes.get(mime) == ctype

    def is_http(self):
        """Return True for *http://* or *https://* URLs."""
        return self.scheme in ("http", "https")

    def is_file(self):
        """Return True for *file://* URLs."""
        return self.scheme == "file"

    def is_directory(self):
        """Return True if current URL represents a directory."""
        return False

    def is_local(self):
        """Return True for local (ie. *file://*) URLs."""
        return self.is_file()

    def add_warning(self, s, tag=None):
        """
        Add a warning string.
        """
        item = (tag, s)
        if (
            item not in self.warnings
            and tag not in self.aggregate.config["ignorewarnings"]
        ):
            self.warnings.append(item)

    def add_info(self, s):
        """
        Add an info string.
        """
        if s not in self.info:
            self.info.append(s)

    def set_cache_url(self):
        """Set the URL to be used for caching."""
        if "AnchorCheck" in self.aggregate.config["enabledplugins"]:
            self.cache_url = self.url
        else:
            # remove anchor from cached target url since we assume
            # URLs with different anchors to have the same content
            self.cache_url = urlutil.urlunsplit(self.urlparts[:4] + [''])
        log.debug(LOG_CHECK, "cache_url '%s'", self.cache_url)

    def check_syntax(self):
        """
        Called before self.check(), this function inspects the
        url syntax. Success enables further checking, failure
        immediately logs this url. Syntax checks must not
        use any network resources.
        """
        log.debug(LOG_CHECK, "checking syntax")
        if self.base_url is None:
            self.base_url = ""
        if not (self.base_url or self.parent_url):
            self.set_result(_("URL is empty"), valid=False)
            return
        try:
            self.build_url()
            self.check_url_warnings()
        except tuple(ExcSyntaxList) as msg:
            self.set_result(str(msg), valid=False)
        else:
            self.set_cache_url()

    def check_url_warnings(self):
        """Check URL name and length."""
        effectiveurl = urlutil.urlunsplit(self.urlparts)
        if self.url != effectiveurl:
            self.add_warning(
                _("Effective URL %(url)r.") % {"url": effectiveurl},
                tag=WARN_URL_EFFECTIVE_URL,
            )
            self.url = effectiveurl
        if len(self.url) > URL_MAX_LENGTH and self.scheme != "data":
            args = dict(len=len(self.url), max=URL_MAX_LENGTH)
            self.add_warning(
                _("URL length %(len)d is longer than %(max)d.") % args,
                tag=WARN_URL_TOO_LONG,
            )

    def build_url(self):
        """
        Construct self.url and self.urlparts out of the given base
        url information self.base_url, self.parent_url and self.base_ref.
        """
        # norm base url - can raise UnicodeError from url.idna_encode()
        base_url, is_idn = url_norm(self.base_url, self.encoding)
        # make url absolute
        if self.base_ref:
            # use base reference as parent url
            if ":" not in self.base_ref:
                # some websites have a relative base reference
                self.base_ref = urljoin(self.parent_url, self.base_ref)
            self.url = urljoin(self.base_ref, base_url)
        elif self.parent_url:
            # strip the parent url anchor
            urlparts = list(urllib.parse.urlsplit(self.parent_url))
            urlparts[4] = ""
            parent_url = urlutil.urlunsplit(urlparts)
            self.url = urljoin(parent_url, base_url)
        else:
            self.url = base_url
        # urljoin can unnorm the url path, so norm it again
        urlparts = list(urllib.parse.urlsplit(self.url))
        if urlparts[2]:
            urlparts[2] = urlutil.collapse_segments(urlparts[2])
            if not urlparts[0].startswith("feed"):
                # restore second / in http[s]:// in wayback path
                urlparts[2] = url_fix_wayback_query(urlparts[2])
        self.url = urlutil.urlunsplit(urlparts)
        self.urlparts = self.build_url_parts(self.url)
        # and unsplit again
        self.url = urlutil.urlunsplit(self.urlparts)

    def build_url_parts(self, url):
        """Set userinfo, host, port and anchor from url and return urlparts.
        Also checks for obfuscated IP addresses.
        """
        split = urllib.parse.urlsplit(url)
        urlparts = list(split)
        # check userinfo@host:port syntax
        self.userinfo, host = urlutil.split_netloc(split.netloc)
        try:
            port = split.port
        except ValueError:
            raise LinkCheckerError(
                _("URL host %(host)r has invalid port") % {"host": host}
            )
        if port is None:
            port = urlutil.default_ports.get(self.scheme, 0)
        if port is None:
            raise LinkCheckerError(
                _("URL host %(host)r has invalid port") % {"host": host}
            )
        self.port = port
        # urllib.parse.SplitResult.hostname is lowercase
        self.host = split.hostname
        if self.scheme in scheme_requires_host:
            if not self.host:
                raise LinkCheckerError(_("URL has empty hostname"))
            self.check_obfuscated_ip()
        if not self.port or self.port == urlutil.default_ports.get(self.scheme):
            host = self.host
        else:
            host = f"{self.host}:{self.port}"
        if self.userinfo:
            urlparts[1] = f"{self.userinfo}@{host}"
        else:
            urlparts[1] = host
        # save anchor for later checking
        self.anchor = split.fragment
        if self.anchor is not None:
            assert isinstance(self.anchor, str), repr(self.anchor)
        return urlparts

    def check_obfuscated_ip(self):
        """Warn if host of this URL is obfuscated IP address."""
        # check if self.host can be an IP address
        # check for obfuscated IP address
        if iputil.is_obfuscated_ip(self.host):
            ips = iputil.resolve_host(self.host)
            if ips:
                self.host = ips[0]
                self.add_warning(
                    _("URL %(url)s has obfuscated IP address %(ip)s")
                    % {"url": self.base_url, "ip": ips[0]},
                    tag=WARN_URL_OBFUSCATED_IP,
                )

    def check(self):
        """Main check function for checking this URL."""
        if self.aggregate.config["trace"]:
            trace.trace_on()
        try:
            self.local_check()
        except OSError:
            # on Unix, ctrl-c can raise
            # error: (4, 'Interrupted system call')
            etype, value = sys.exc_info()[:2]
            if etype == errno.EINTR:
                raise KeyboardInterrupt(value)
            else:
                raise

    def local_check(self):
        """Local check function can be overridden in subclasses."""
        log.debug(LOG_CHECK, "Checking %s", self)
        # strict extern URLs should not be checked
        assert not self.extern[1], 'checking strict extern URL'
        # check connection
        log.debug(LOG_CHECK, "checking connection")
        try:
            self.check_connection()
            self.set_content_type()
            self.add_size_info()
            self.aggregate.plugin_manager.run_connection_plugins(self)
        except tuple(ExcList) as exc:
            value = self.handle_exception()
            # make nicer error msg for unknown hosts
            if isinstance(exc, socket.error) and exc.args[0] == -2:
                value = _('Hostname not found')
            elif isinstance(exc, UnicodeError):
                # idna.encode(host) failed
                value = _('Bad hostname %(host)r: %(msg)s') % {
                    'host': self.host,
                    'msg': value,
                }
            self.set_result(value, valid=False)

    def check_content(self):
        """Check content of URL.
        @return: True if content can be parsed, else False
        """
        if self.do_check_content and self.valid:
            # check content and recursion
            try:
                if self.can_get_content():
                    self.aggregate.plugin_manager.run_content_plugins(self)
                if self.allows_recursion():
                    return True
            except tuple(ExcList):
                value = self.handle_exception()
                self.add_warning(
                    _("could not get content: %(msg)s") % {"msg": value},
                    tag=WARN_URL_ERROR_GETTING_CONTENT,
                )
        return False

    def close_connection(self):
        """
        Close an opened url connection.
        """
        if self.url_connection is None:
            # no connection is open
            return
        try:
            self.url_connection.close()
        except Exception:
            # ignore close errors
            pass
        self.url_connection = None

    def handle_exception(self):
        """
        An exception occurred. Log it and set the cache flag.
        """
        etype, evalue = sys.exc_info()[:2]
        log.debug(
            LOG_CHECK, "Error in %s: %s %s", self.url, etype, evalue, exception=True
        )
        # note: etype must be the exact class, not a subclass
        if (
            (etype in ExcNoCacheList)
            or (etype == socket.error and evalue.args[0] == errno.EBADF)
            or not evalue
        ):
            # EBADF occurs when operating on an already socket
            self.caching = False
        # format message "<exception name>: <error message>"
        errmsg = etype.__name__
        if evalue:
            errmsg += f": {evalue}"
        # limit length to 240
        return strformat.limit(errmsg, length=240)

    def check_connection(self):
        """
        The basic connection check uses urlopen to initialize
        a connection object.
        """
        self.url_connection = urlopen(self.url)

    def add_size_info(self):
        """Set size of URL content (if any)..
        Should be overridden in subclasses."""
        maxbytes = self.aggregate.config["maxfilesizedownload"]
        if self.size > maxbytes:
            self.add_warning(
                _("Content size %(size)s is larger than %(maxbytes)s.")
                % dict(
                    size=strformat.strsize(self.size),
                    maxbytes=strformat.strsize(maxbytes),
                ),
                tag=WARN_URL_CONTENT_SIZE_TOO_LARGE,
            )

    def allows_simple_recursion(self):
        """Check recursion level and extern status."""
        rec_level = self.aggregate.config["recursionlevel"]
        if rec_level >= 0 and self.recursion_level >= rec_level:
            log.debug(LOG_CHECK, "... no, maximum recursion level reached.")
            return False
        if self.extern[0]:
            log.debug(LOG_CHECK, "... no, extern.")
            return False
        return True

    def allows_recursion(self):
        """
        Return True iff we can recurse into the url's content.
        """
        log.debug(LOG_CHECK, "checking recursion of %r ...", self.url)
        if not self.valid:
            log.debug(LOG_CHECK, "... no, invalid.")
            return False
        if not self.can_get_content():
            log.debug(LOG_CHECK, "... no, cannot get content.")
            return False
        if not self.allows_simple_recursion():
            return False
        if self.size > self.aggregate.config["maxfilesizeparse"]:
            log.debug(LOG_CHECK, "... no, maximum parse size.")
            return False
        if not self.is_parseable():
            log.debug(LOG_CHECK, "... no, not parseable.")
            return False
        if not self.content_allows_robots():
            log.debug(LOG_CHECK, "... no, robots.")
            return False
        log.debug(LOG_CHECK, "... yes, recursion.")
        return True

    def content_allows_robots(self):
        """Returns True: only check robots.txt on HTTP links."""
        return True

    def set_extern(self, url):
        """
        Match URL against extern and intern link patterns. If no pattern
        matches the URL is extern. Sets self.extern to a tuple (bool,
        bool) with content (is_extern, is_strict).

        @return: None
        """
        if self.extern:
            return
        if not url:
            self.extern = (1, 1)
            return
        for entry in self.aggregate.config["externlinks"]:
            match = entry['pattern'].search(url)
            if (entry['negate'] and not match) or (match and not entry['negate']):
                log.debug(LOG_CHECK, "Extern URL %r", url)
                self.extern = (1, entry['strict'])
                return
        for entry in self.aggregate.config["internlinks"]:
            match = entry['pattern'].search(url)
            if (entry['negate'] and not match) or (match and not entry['negate']):
                log.debug(LOG_CHECK, "Intern URL %r", url)
                self.extern = (0, 0)
                return
        if self.aggregate.config['checkextern']:
            self.extern = (1, 0)
        else:
            self.extern = (1, 1)

    def set_content_type(self):
        """Set content MIME type.
        Should be overridden in subclasses."""
        pass

    def can_get_content(self):
        """Indicate whether url get_content() can be called."""
        return self.size <= self.aggregate.config["maxfilesizedownload"]

    def download_content(self):
        log.debug(LOG_CHECK, "Get content of %r", self.url)
        t = time.time()
        content = self.read_content()
        self.size = len(content)
        self.dltime = time.time() - t
        if self.size == 0:
            self.add_warning(_("Content size is zero."), tag=WARN_URL_CONTENT_SIZE_ZERO)
        else:
            self.aggregate.add_downloaded_bytes(self.size)
        return content

    def get_soup(self):
        if self.soup is None:
            self.get_content()
        return self.soup

    def get_raw_content(self):
        if self.data is None:
            self.data = self.download_content()
        return self.data

    def get_content(self, encoding=None):
        if self.text is None:
            self.get_raw_content()
            self.soup = htmlsoup.make_soup(self.data, encoding)
            # Sometimes soup.original_encoding is None!  Better mangled text
            # than an internal crash, eh?  ISO-8859-1 is a safe fallback in the
            # sense that any binary blob can be decoded, it'll never cause a
            # UnicodeDecodeError.
            log.debug(
                LOG_CHECK, "Beautiful Soup detected %s", self.soup.original_encoding
            )
            self.content_encoding = self.soup.original_encoding or 'ISO-8859-1'
            log.debug(LOG_CHECK, "Content encoding %s", self.content_encoding)
            self.text = self.data.decode(self.content_encoding)
        return self.text

    def read_content(self):
        """Return data for this URL. Can be overridden in subclasses."""
        buf = BytesIO()
        data = self.read_content_chunk()
        while data:
            if buf.tell() + len(data) > self.aggregate.config["maxfilesizedownload"]:
                raise LinkCheckerError(_("File size too large"))
            buf.write(data)
            data = self.read_content_chunk()
        return buf.getvalue()

    def read_content_chunk(self):
        """Read one chunk of content from this URL.
        Precondition: url_connection is an opened URL.
        """
        return self.url_connection.read(self.ReadChunkBytes)

    def get_user_password(self):
        """Get tuple (user, password) from configured authentication.
        Both user and password can be None.
        """
        if self.userinfo:
            # URL itself has authentication info
            split = urllib.parse.urlsplit(self.url)
            return (split.username, split.password)
        return self.aggregate.config.get_user_password(self.url)

    def add_url(self, url, line=0, column=0, page=0, name="", base=None, parent=None):
        """Add new URL to queue."""
        if base:
            base_ref = urlutil.url_norm(base, encoding=self.content_encoding)[0]
        else:
            base_ref = None
        url_data = get_url_from(
            url,
            self.recursion_level + 1,
            self.aggregate,
            parent_url=self.url if parent is None else parent,
            base_ref=base_ref,
            line=line,
            column=column,
            page=page,
            name=name,
            parent_content_type=self.content_type,
            url_encoding=self.content_encoding,
        )
        self.aggregate.urlqueue.put(url_data)

    def serialized(self, sep=os.linesep):
        """
        Return serialized url check data as unicode string.
        """
        return sep.join(
            [
                "%s link" % self.scheme,
                "base_url=%r" % self.base_url,
                "parent_url=%r" % self.parent_url,
                "base_ref=%r" % self.base_ref,
                "recursion_level=%d" % self.recursion_level,
                "url_connection=%s" % self.url_connection,
                "line=%s" % self.line,
                "column=%s" % self.column,
                "page=%d" % self.page,
                "name=%r" % self.name,
                "anchor=%r" % self.anchor,
                "cache_url=%s" % self.cache_url,
            ]
        )

    def get_intern_pattern(self, url=None):
        """Get pattern for intern URL matching.

        @param url: the URL to set intern pattern for, else self.url
        @type url: unicode or None
        @return: non-empty regex pattern or None
        @rtype: String or None
        """
        return None

    def add_intern_pattern(self, url=None):
        """Add intern URL regex to config."""
        try:
            pat = self.get_intern_pattern(url=url)
            if pat:
                log.debug(LOG_CHECK, "Add intern pattern %r", pat)
                self.aggregate.config['internlinks'].append(get_link_pat(pat))
        except UnicodeError as msg:
            res = _("URL has unparsable domain name: %(domain)s") % {"domain": msg}
            self.set_result(res, valid=False)

    def __str__(self):
        """
        Get URL info.

        @return: URL info
        @rtype: unicode
        """
        return self.serialized()

    def __bytes__(self):
        """
        Get URL info.

        @return: URL info, encoded with the output logger encoding
        @rtype: string
        """
        s = str(self)
        return self.aggregate.config['logger'].encode(s)

    def __repr__(self):
        """
        Get URL info.

        @return: URL info
        @rtype: unicode
        """
        return f'<{self.serialized(sep=", ")}>'

    def to_wire_dict(self):
        """Return a simplified transport object for logging and caching.

        The transport object must contain these attributes:

        - url_data.valid: bool
          Indicates if URL is valid
        - url_data.result: unicode
          Result string
        - url_data.warnings: list of tuples (tag, warning message)
          List of tagged warnings for this URL.
        - url_data.name: unicode string or None
          name of URL (eg. filename or link name)
        - url_data.parent_url: unicode or None
          Parent URL
        - url_data.base_ref: unicode
          HTML base reference URL of parent
        - url_data.url: unicode
          Fully qualified URL.
        - url_data.domain: unicode
          URL domain part.
        - url_data.checktime: int
          Number of seconds needed to check this link, default: zero.
        - url_data.dltime: int
          Number of seconds needed to download URL content, default: -1
        - url_data.size: int
          Size of downloaded URL content, default: -1
        - url_data.info: list of unicode
          Additional information about this URL.
        - url_data.line: int
          Line number of this URL at parent document, or None
        - url_data.column: int
          Column number of this URL at parent document, or None
        - url_data.page: int
          Page number of this URL at parent document, or -1
        - url_data.cache_url: unicode
          Cache url for this URL.
        - url_data.content_type: unicode
          MIME content type for URL content.
        - url_data.level: int
          Recursion level until reaching this URL from start URL
        - url_data.last_modified: datetime
          Last modification date of retrieved page (or None).
        """
        return dict(
            valid=self.valid,
            extern=self.extern[0],
            result=self.result,
            warnings=self.warnings[:],
            name=self.name or "",
            title=self.get_title(),
            parent_url=self.parent_url or "",
            base_ref=self.base_ref or "",
            base_url=self.base_url or "",
            url=self.url or "",
            domain=(self.urlparts[1] if self.urlparts else ""),
            checktime=self.checktime,
            dltime=self.dltime,
            size=self.size,
            info=self.info,
            line=self.line,
            column=self.column,
            page=self.page,
            cache_url=self.cache_url,
            content_type=self.content_type,
            level=self.recursion_level,
            modified=self.modified,
        )

    def to_wire(self):
        """Return compact UrlData object with information from to_wire_dict().
        """
        return CompactUrlData(self.to_wire_dict())


urlDataAttr = [
    'valid',
    'extern',
    'result',
    'warnings',
    'name',
    'title',
    'parent_url',
    'base_ref',
    'base_url',
    'url',
    'domain',
    'checktime',
    'dltime',
    'size',
    'info',
    'modified',
    'line',
    'column',
    'page',
    'cache_url',
    'content_type',
    'level',
]


class CompactUrlData:
    """Store selected UrlData attributes in slots to minimize memory usage."""

    __slots__ = urlDataAttr

    def __init__(self, wired_url_data):
        '''Set all attributes according to the dictionary wired_url_data'''
        for attr in urlDataAttr:
            setattr(self, attr, wired_url_data[attr])



# Copyright (C) 2000-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
A failures logger.
"""

import os

from . import _Logger
from .. import log, LOG_CHECK
from ..configuration import get_user_data


class FailuresLogger(_Logger):
    """
    Updates a list of failed links. If a link already on the list
    is found to be working, it is removed. After n days
    we only have links on the list which failed within those n days.
    """

    LoggerName = "failures"

    LoggerArgs = {
        "filename": os.path.join(get_user_data(), "failures"),
    }

    def __init__(self, **kwargs):
        """Initialize with old failures data (if found)."""
        blacklist = os.path.join(get_user_data(), "blacklist")
        if os.path.isfile(blacklist):
            self.LoggerArgs["filename"] = blacklist
            log.warn(
                LOG_CHECK,
                _("%(blacklist)s file is deprecated please rename to failures")
                % {"blacklist": blacklist}
            )
        args = self.get_args(kwargs)
        super().__init__(**args)
        self.init_fileoutput(args)
        self.failures = {}
        if self.filename is not None and os.path.exists(self.filename):
            self.read_failures()

    def comment(self, s, **args):
        """
        Write nothing.
        """
        pass

    def log_url(self, url_data):
        """
        Add invalid url to failures, delete valid url from failures.
        """
        key = (url_data.parent_url, url_data.cache_url)
        key = repr(key)
        if key in self.failures:
            if url_data.valid:
                del self.failures[key]
            else:
                self.failures[key] += 1
        else:
            if not url_data.valid:
                self.failures[key] = 1

    def end_output(self, **kwargs):
        """
        Write failures file.
        """
        self.write_failures()

    def read_failures(self):
        """
        Read a previously stored failures from file fd.
        """
        with open(self.filename, encoding=self.output_encoding,
                  errors=self.codec_errors) as fd:
            for line in fd:
                line = line.rstrip()
                if line.startswith('#') or not line:
                    continue
                value, key = line.split(None, 1)
                key = key.strip('"')
                if not key.startswith('('):
                    log.critical(
                        LOG_CHECK,
                        _("invalid line starting with '%(linestart)s' in %(failures)s")
                        % {"linestart": line[:12], "failures": self.filename}
                    )
                    raise SystemExit(2)
                self.failures[key] = int(value)

    def write_failures(self):
        """
        Write the failures file.
        """
        oldmask = os.umask(0o077)
        for key, value in self.failures.items():
            self.write("%d %s%s" % (value, repr(key), os.linesep))
        self.close_fileoutput()
        # restore umask
        os.umask(oldmask)



# Copyright (C) 2000-2018 Petr Dlouhy
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
HTML parser implemented using Beautiful Soup and html.parser.
"""

import warnings

warnings.filterwarnings(
    "ignore",
    message="The soupsieve package is not installed. CSS selectors cannot be used.",
    category=UserWarning,
    module="bs4",
)

import bs4

# bs4 4.9.1 introduced MarkupResemblesLocatorWarning
hasattr(bs4, "MarkupResemblesLocatorWarning") and warnings.simplefilter(
    'ignore', bs4.MarkupResemblesLocatorWarning
)

# bs4 4.11.0 introduced builder.XMLParsedAsHTMLWarning
hasattr(bs4.builder, "XMLParsedAsHTMLWarning") and warnings.simplefilter(
    'ignore', bs4.builder.XMLParsedAsHTMLWarning
)


def make_soup(markup, from_encoding=None):
    return bs4.BeautifulSoup(
        markup, "html.parser", from_encoding=from_encoding, multi_valued_attributes=None
    )



# Copyright (C) 2005-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Intern URL pattern support.
"""

import re
import urllib.parse

from . import urlbase, absolute_url
from .. import url as urlutil


def get_intern_pattern(url):
    """Return intern pattern for given URL. Redirections to the same
    domain with or without "www." prepended are allowed."""
    parts = urllib.parse.urlsplit(url)
    scheme = parts[0].lower()
    domain = parts[1].lower()
    domain, is_idn = urlutil.idna_encode(domain)
    # allow redirection www.example.com -> example.com and vice versa
    if domain.startswith('www.'):
        domain = domain[4:]
    if not (domain and scheme):
        return None
    path = urlutil.splitparams(parts[2])[0]
    segments = path.split('/')[:-1]
    path = "/".join(segments)
    if url.endswith('/'):
        path += '/'
    args = list(re.escape(x) for x in (scheme, domain, path))
    if args[0] in ('http', 'https'):
        args[0] = 'https?'
    args[1] = r"(www\.|)%s" % args[1]
    return "^%s://%s%s" % tuple(args)


class InternPatternUrl(urlbase.UrlBase):
    """Class supporting an intern URL pattern."""

    def get_intern_pattern(self, url=None):
        """
        Get pattern for intern URL matching.

        @return non-empty regex pattern or None
        @rtype String or None
        """
        if url is None:
            url = absolute_url(self.base_url, self.base_ref, self.parent_url)
        if not url:
            return None
        return get_intern_pattern(url)



# Copyright (C) 2005-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
File and path utilities.
"""

import os
import locale
import stat
import tempfile
import importlib
from functools import lru_cache


def has_module(name, without_error=True):
    """Test if given module can be imported.
    @param without_error: True if module must not throw any errors when importing
    @return: flag if import is successful
    @rtype: bool
    """
    try:
        importlib.import_module(name)
        return True
    except ImportError:
        return False
    except Exception:
        # some modules raise errors when intitializing
        return not without_error


def get_mtime(filename):
    """Return modification time of filename or zero on errors."""
    try:
        return os.path.getmtime(filename)
    except os.error:
        return 0


def get_size(filename):
    """Return file size in Bytes, or -1 on error."""
    try:
        return os.path.getsize(filename)
    except os.error:
        return -1


# http://developer.gnome.org/doc/API/2.0/glib/glib-running.html
if "G_FILENAME_ENCODING" in os.environ:
    FSCODING = os.environ["G_FILENAME_ENCODING"].split(",")[0]
    if FSCODING == "@locale":
        FSCODING = locale.getpreferredencoding()
elif "G_BROKEN_FILENAMES" in os.environ:
    FSCODING = locale.getpreferredencoding()
else:
    FSCODING = "utf-8"


def path_safe(path):
    """Ensure path string is compatible with the platform file system encoding."""
    if path and not os.path.supports_unicode_filenames:
        path = path.encode(FSCODING, "replace").decode(FSCODING)
    return path


def get_temp_file(mode='r', **kwargs):
    """Return tuple (open file object, filename) pointing to a temporary
    file."""
    fd, filename = tempfile.mkstemp(**kwargs)
    return os.fdopen(fd, mode), filename


def is_tty(fp):
    """Check if is a file object pointing to a TTY."""
    return hasattr(fp, "isatty") and fp.isatty()


@lru_cache(128)
def is_readable(filename):
    """Check if file is a regular file and is readable."""
    return os.path.isfile(filename) and os.access(filename, os.R_OK)


def is_accessable_by_others(filename):
    """Check if file is group or world accessible."""
    mode = os.stat(filename)[stat.ST_MODE]
    return mode & (stat.S_IRWXG | stat.S_IRWXO)


def is_writable_by_others(filename):
    """Check if file or directory is world writable."""
    mode = os.stat(filename)[stat.ST_MODE]
    return mode & stat.S_IWOTH



# Copyright (C) 2000-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Functions for parsing and matching URL strings.
"""

import os
import re
import urllib.parse

for scheme in ('ldap', 'irc'):
    if scheme not in urllib.parse.uses_netloc:
        urllib.parse.uses_netloc.append(scheme)

# The character set to encode non-ASCII characters in a URL. See also
# http://tools.ietf.org/html/rfc2396#section-2.1
# Note that the encoding is not really specified, but most browsers
# encode in UTF-8 when no encoding is specified by the HTTP headers,
# else they use the page encoding for followed link. See also
# http://code.google.com/p/browsersec/wiki/Part1#Unicode_in_URLs
url_encoding = "utf-8"

default_ports = {
    'http': 80,
    'https': 443,
    'nntps': 563,
    'ftp': 21,
}

# adapted from David Wheelers "Secure Programming for Linux and Unix HOWTO"
# http://www.dwheeler.com/secure-programs/Secure-Programs-HOWTO/\
# filter-html.html#VALIDATING-URIS
_path = r"\-\_\.\!\~\*\'\(\),"
_hex_safe = r"2-9a-f"
_hex_full = r"0-9a-f"
_part = r"([a-z0-9][-a-z0-9]{0,61}|[a-z])"
_safe_char = (
    fr"([a-z0-9{_path}\+]|"
    fr"(%[{_hex_safe}][{_hex_full}]))"
)
_safe_scheme_pattern = r"(https?|ftp)"
_safe_domain_pattern = fr"({_part}(\.{_part})*\.?)"
_safe_host_pattern = fr"{_safe_domain_pattern}(:(80|8080|8000|443))?"
_safe_path_pattern = (
    fr"((/([a-z0-9{_path}]|"
    fr"(%[{_hex_safe}][{_hex_full}]))+)*/?)"
)
_safe_fragment_pattern = fr"{_safe_char}*"
_safe_cgi = fr"{_safe_char}+(=({_safe_char}|/)+)?"
_safe_query_pattern = fr"({_safe_cgi}(&{_safe_cgi})*)?"
_safe_param_pattern = fr"({_safe_cgi}(;{_safe_cgi})*)?"
safe_url_pattern = r"{}://{}{}(#{})?".format(
    _safe_scheme_pattern,
    _safe_host_pattern,
    _safe_path_pattern,
    _safe_fragment_pattern,
)

is_safe_url = re.compile(f"(?i)^{safe_url_pattern}$").match
is_safe_domain = re.compile(f"(?i)^{_safe_domain_pattern}$").match


# snatched form urlparse.py
def splitparams(path):
    """Split off parameter part from path.
    Returns tuple (path-without-param, param)
    """
    if '/' in path:
        i = path.find(';', path.rfind('/'))
    else:
        i = path.find(';')
    if i < 0:
        return path, ''
    return path[:i], path[i + 1:]


def is_numeric_port(portstr):
    """return: integer port (== True) iff portstr is a valid port number,
           False otherwise
    """
    if portstr.isdigit():
        port = int(portstr)
        # 65536 == 2**16
        if 0 < port < 65536:
            return port
    return False


def parse_qsl(qs, encoding, keep_blank_values=0, strict_parsing=0):
    """Parse a query given as a string argument.

    @param qs: URL-encoded query string to be parsed
    @type qs: string
    @param keep_blank_values: flag indicating whether blank values in
        URL encoded queries should be treated as blank strings.  A
        true value indicates that blanks should be retained as blank
        strings.  The default false value indicates that blank values
        are to be ignored and treated as if they were  not included.
    @type keep_blank_values: bool
    @param strict_parsing: flag indicating what to do with parsing errors. If
        false (the default), errors are silently ignored. If true,
        errors raise a ValueError exception.
    @type strict_parsing: bool
    @returns: list of triples (key, value, separator) where key and value
      are the split CGI parameter and separator the used separator
      for this CGI parameter which is either a semicolon or an ampersand
    @rtype: list of triples
    """
    pairs = []
    name_value_amp = qs.split('&')
    for name_value in name_value_amp:
        if ';' in name_value:
            pairs.extend([x, ';'] for x in name_value.split(';'))
            pairs[-1][1] = '&'
        else:
            pairs.append([name_value, '&'])
    pairs[-1][1] = ''
    r = []
    for name_value, sep in pairs:
        nv = name_value.split('=', 1)
        if len(nv) != 2:
            if strict_parsing:
                raise ValueError("bad query field: %r" % name_value)
            elif len(nv) == 1:
                # None value indicates missing equal sign
                nv = (nv[0], None)
            else:
                continue
        if nv[1] or keep_blank_values:
            name = urllib.parse.unquote(nv[0].replace('+', ' '), encoding=encoding)
            if nv[1]:
                value = urllib.parse.unquote(nv[1].replace('+', ' '), encoding=encoding)
            else:
                value = nv[1]
            r.append((name, value, sep))
    return r


def idna_encode(host):
    """Encode hostname as internationalized domain name (IDN) according
    to RFC 3490.
    @raise: UnicodeError if hostname is not properly IDN encoded.
    """
    if host:
        try:
            host.encode('ascii')
            return host, False
        except UnicodeError:
            uhost = host.encode('idna').decode('ascii')
            return uhost, uhost != host
    return host, False


def split_netloc(netloc):
    """Separate userinfo from host in urllib.parse.SplitResult.netloc.
    Originated as urllib.parse._splituser().
    """
    userinfo, delim, hostport = netloc.rpartition('@')
    return (userinfo if delim else None), hostport


def url_fix_host(urlparts, encoding):
    """Unquote and fix hostname. Returns is_idn."""
    if not urlparts[1]:
        urlparts[2] = urllib.parse.unquote(urlparts[2], encoding=encoding)
        return False
    userpass, hostport = split_netloc(urlparts[1])
    if userpass:
        userpass = urllib.parse.unquote(userpass, encoding=encoding)
    netloc, is_idn = idna_encode(
        urllib.parse.unquote(hostport, encoding=encoding).lower()
    )
    # a leading backslash in path causes urlsplit() to add the
    # path components up to the first slash to host
    # try to find this case...
    i = netloc.find("\\")
    if i != -1:
        # ...and fix it by prepending the misplaced components to the path
        comps = netloc[i:]  # note: still has leading backslash
        if not urlparts[2] or urlparts[2] == '/':
            urlparts[2] = comps
        else:
            urlparts[2] = "{}{}".format(
                comps,
                urllib.parse.unquote(urlparts[2], encoding=encoding),
            )
        netloc = netloc[:i]
    else:
        # a leading ? in path causes urlsplit() to add the query to the
        # host name
        i = netloc.find("?")
        if i != -1:
            netloc, urlparts[3] = netloc.split('?', 1)
        # path
        urlparts[2] = urllib.parse.unquote(urlparts[2], encoding=encoding)
    if userpass:
        # append AT for easy concatenation
        userpass += "@"
    else:
        userpass = ""
    if urlparts[0] in default_ports:
        dport = default_ports[urlparts[0]]
        host, port = splitport(netloc, port=dport)
        if host.endswith("."):
            host = host[:-1]
        if port != dport:
            host = f"{host}:{port}"
        netloc = host
    urlparts[1] = userpass + netloc
    return is_idn


def url_fix_mailto_urlsplit(urlparts):
    """Split query part of mailto url if found."""
    sep = "?"
    if sep in urlparts[2]:
        urlparts[2], urlparts[3] = urlparts[2].split(sep, 1)


# wayback urls include in the path http[s]://. By default the
# tidying mechanism in linkchecker encodes the : and deletes the second slash
# This function reverses these corrections. This function expects only the
# path section of the URL as input.
wayback_regex = re.compile(r'(https?)(\%3A/|:/)')


def url_fix_wayback_query(path):
    return wayback_regex.sub(r'\1://', path)


def url_parse_query(query, encoding):
    """Parse and re-join the given CGI query."""
    # if ? is in the query, split it off, seen at msdn.microsoft.com
    append = ""
    while '?' in query:
        query, rest = query.rsplit('?', 1)
        append = '?' + url_parse_query(rest, encoding=encoding) + append
    f = []
    for k, v, sep in parse_qsl(query, keep_blank_values=True, encoding=encoding):
        k = urllib.parse.quote(k, safe='/-:,;')
        if v:
            v = urllib.parse.quote(v, safe='/-:,;')
            f.append(f"{k}={v}{sep}")
        elif v is None:
            f.append(f"{k}{sep}")
        else:
            # some sites do not work when the equal sign is missing
            f.append(f"{k}={sep}")
    return ''.join(f) + append


def urlunsplit(urlparts):
    """Same as urllib.parse.urlunsplit but with extra UNC path handling
    for Windows OS."""
    res = urllib.parse.urlunsplit(urlparts)
    if os.name == 'nt' and urlparts[0] == 'file' and '|' not in urlparts[2]:
        # UNC paths must have 4 slashes: 'file:////server/path'
        # Depending on the path in urlparts[2], urllib.parse.urlunsplit()
        # left only two or three slashes. This is fixed below
        repl = 'file://' if urlparts[2].startswith('//') else 'file:/'
        res = res.replace('file:', repl)
    return res


def url_norm(url, encoding):
    """Normalize the given URL which must be quoted. Supports unicode
    hostnames (IDNA encoding) according to RFC 3490.

    @return: (normed url, idna flag)
    @rtype: tuple of length two
    """
    urlparts = list(urllib.parse.urlsplit(url))
    # scheme
    urlparts[0] = urllib.parse.unquote(urlparts[0], encoding=encoding).lower()
    # mailto: urlsplit is broken
    if urlparts[0] == 'mailto':
        url_fix_mailto_urlsplit(urlparts)
    # host (with path or query side effects)
    is_idn = url_fix_host(urlparts, encoding)
    # query
    urlparts[3] = url_parse_query(urlparts[3], encoding=encoding)
    if urlparts[0] in urllib.parse.uses_relative:
        # URL has a hierarchical path we should norm
        if not urlparts[2]:
            # Empty path is allowed if both query and fragment are also empty.
            # Note that in relative links, urlparts[0] might be empty.
            # In this case, do not make any assumptions.
            if urlparts[0] and (urlparts[3] or urlparts[4]):
                urlparts[2] = '/'
        else:
            # fix redundant path parts
            urlparts[2] = collapse_segments(urlparts[2])
    # anchor
    urlparts[4] = urllib.parse.unquote(urlparts[4], encoding=encoding)
    # quote parts again
    urlparts[0] = urllib.parse.quote(urlparts[0])  # scheme
    urlparts[1] = urllib.parse.quote(urlparts[1], safe='@:')  # host
    urlparts[2] = urllib.parse.quote(urlparts[2], safe=_nopathquote_chars)  # path
    if not urlparts[0].startswith("feed"):
        # unencode colon in http[s]:// in wayback path
        urlparts[2] = url_fix_wayback_query(urlparts[2])
    urlparts[4] = urllib.parse.quote(urlparts[4], safe="!$&'()*+,-./;=?@_~")  # anchor
    res = urlunsplit(urlparts)
    if url.endswith('#') and not urlparts[4]:
        # re-append trailing empty fragment
        res += '#'
    return (res, is_idn)


_slashes_ro = re.compile(r"/+")
_thisdir_ro = re.compile(r"^\./")
_samedir_ro = re.compile(r"/\./|/\.$")
_parentdir_ro = re.compile(r"^/(\.\./)+|/(?!\.\./)[^/]+/\.\.(/|$)")
_relparentdir_ro = re.compile(r"^(?!\.\./)[^/]+/\.\.(/|$)")


def collapse_segments(path):
    """Remove all redundant segments from the given URL path.
    Precondition: path is an unquoted url path"""
    # replace backslashes
    # note: this is _against_ the specification (which would require
    # backslashes to be left alone, and finally quoted with '%5C')
    # But replacing has several positive effects:
    # - Prevents path attacks on Windows systems (using \.. parent refs)
    # - Fixes bad URLs where users used backslashes instead of slashes.
    #   This is a far more probable case than users having an intentional
    #   backslash in the path name.
    path = path.replace('\\', '/')
    # shrink multiple slashes to one slash
    path = _slashes_ro.sub("/", path)
    # collapse redundant path segments
    path = _thisdir_ro.sub("", path)
    path = _samedir_ro.sub("/", path)
    # collapse parent path segments
    # note: here we exploit the fact that the replacements happen
    # to be from left to right (see also _parentdir_ro above)
    newpath = _parentdir_ro.sub("/", path)
    while newpath != path:
        path = newpath
        newpath = _parentdir_ro.sub("/", path)
    # collapse parent path segments of relative paths
    # (ie. without leading slash)
    newpath = _relparentdir_ro.sub("", path)
    while newpath != path:
        path = newpath
        newpath = _relparentdir_ro.sub("", path)
    return path


url_is_absolute = re.compile(r"^[-\.a-z]+:", re.I).match


def url_quote(url, encoding):
    """Quote given URL."""
    if not url_is_absolute(url):
        return document_quote(url)
    urlparts = list(urllib.parse.urlsplit(url))
    urlparts[0] = urllib.parse.quote(urlparts[0])  # scheme
    urlparts[1] = urllib.parse.quote(urlparts[1], safe=':')  # host
    urlparts[2] = urllib.parse.quote(urlparts[2], safe='/=,')  # path
    urlparts[3] = urllib.parse.quote(urlparts[3], safe='&=,')  # query
    f = []
    for k, v, sep in parse_qsl(
        urlparts[3], encoding=encoding, keep_blank_values=True
    ):  # query
        k = urllib.parse.quote(k, safe='/-:,;')
        if v:
            v = urllib.parse.quote(v, safe='/-:,;')
            f.append(f"{k}={v}{sep}")
        else:
            f.append(f"{k}{sep}")
    urlparts[3] = ''.join(f)
    urlparts[4] = urllib.parse.quote(urlparts[4])  # anchor
    return urlunsplit(urlparts)


def document_quote(document):
    """Quote given document."""
    doc, delim, query = document.rpartition('?')
    if not delim:
        doc = document
        query = None
    doc = urllib.parse.quote(doc, safe='/=,')
    if query:
        return f"{doc}?{query}"
    return doc


_nopathquote_chars = "-;/=,~*+()@!"
if os.name == 'nt':
    _nopathquote_chars += "|"
_safe_url_chars = re.escape(_nopathquote_chars + "_:.&#%?[]!") + "a-zA-Z0-9"
_safe_url_chars_ro = re.compile(fr"^[{_safe_url_chars}]*$")


def url_needs_quoting(url):
    """Check if url needs percent quoting. Note that the method does
    only check basic character sets, and not any other syntax.
    The URL might still be syntactically incorrect even when
    it is properly quoted.
    """
    if url.rstrip() != url:
        # handle trailing whitespace as a special case
        # since '$' matches immediately before a end-of-line
        return True
    return not _safe_url_chars_ro.match(url)


def splitport(host, port=0):
    """Split optional port number from host. If host has no port number,
    the given default port is returned.

    @param host: host name
    @type host: string
    @param port: the port number (default 0)
    @type port: int

    @return: tuple of (host, port)
    @rtype: tuple of (string, int)
    """
    if ":" in host:
        shost, sport = host.split(":", 1)
        iport = is_numeric_port(sport)
        if iport:
            host, port = shost, iport
        elif not sport:
            # empty port, ie. the host was "hostname:"
            host = shost
        else:
            # For an invalid non-empty port leave the host name as is
            pass
    return host, port



# Copyright (C) 2000-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Check HTML pages for broken links. This is the commandline
client. Run this file with the -h option to see how it's done.
"""


import os
import pprint
import signal
import sys
import traceback

from .arg_parser import ArgParser
from .setup_config import setup_config

from .. import configuration
from .. import fileutil
from .. import log
from .. import logconf
from .. import LinkCheckerError
from ..cmdline import aggregate_url, print_usage
from ..director import console, check_urls, get_aggregate
from ..logconf import LOG_CHECK, LOG_CMDLINE, LOG_THREAD
from ..strformat import stripurl


def drop_privileges():
    """Make sure to drop root privileges on POSIX systems."""
    if os.name != 'posix':
        return
    if os.geteuid() == 0:
        log.warn(
            LOG_CHECK,
            _(
                "Running as root user; "
                "dropping privileges by changing user to nobody."
            ),
        )
        import pwd

        os.seteuid(pwd.getpwnam('nobody')[3])


def linkchecker():
    if hasattr(signal, "SIGUSR1"):
        # install SIGUSR1 handler
        from ..decorators import signal_handler

        @signal_handler(signal.SIGUSR1)
        def print_threadstacks(sig, frame):
            """Print stack traces of all running threads."""
            log.warn(LOG_THREAD, "*** STACKTRACE START ***")
            for threadId, stack in sys._current_frames().items():
                log.warn(LOG_THREAD, "# ThreadID: %s" % threadId)
                for filename, lineno, name, line in traceback.extract_stack(stack):
                    log.warn(
                        LOG_THREAD,
                        'File: "%s", line %d, in %s' % (filename, lineno, name)
                    )
                    line = line.strip()
                    if line:
                        log.warn(LOG_THREAD, "  %s" % line)
            log.warn(LOG_THREAD, "*** STACKTRACE END ***")

    logconf.init_log_config()

    # optional modules
    has_argcomplete = fileutil.has_module("argcomplete")
    has_profile = fileutil.has_module("yappi")
    has_meliae = fileutil.has_module("meliae")

    # default profiling filename
    _profile = "linkchecker.prof"

    def read_stdin_urls():
        """Read list of URLs, separated by white-space, from stdin."""
        num = 0
        while True:
            lines = sys.stdin.readlines(8 * 1024)
            if not lines:
                break
            for line in lines:
                for url in line.split():
                    num += 1
                    if num % 10000 == 0:
                        log.info(LOG_CMDLINE, "Read %d URLs from stdin", num)
                    yield url

    # instantiate command line option parser
    argparser = ArgParser()

    # build a config object for this check session
    config = configuration.Configuration()
    config.set_status_logger(console.StatusLogger())

    # ================= auto completion =====================
    if has_argcomplete:
        import argcomplete

        argcomplete.autocomplete(argparser)

    # read and parse command line options and arguments
    options = argparser.parse_args()
    # initialize logging
    if options.debug:
        allowed_debugs = logconf.lognames.keys()
        for _name in options.debug:
            if _name not in allowed_debugs:
                print_usage(_("Invalid debug level %(level)r") % {"level": _name})
        logconf.set_debug(options.debug)
    elif options.quiet:
        logconf.reset_loglevel()
    log.debug(
        LOG_CMDLINE,
        _("Python %(version)s on %(platform)s")
        % {"version": sys.version, "platform": sys.platform},
    )
    # read configuration files
    try:
        files = []
        if options.configfile:
            path = configuration.normpath(options.configfile)
            if os.path.isfile(path):
                files.append(path)
            else:
                log.warn(
                    LOG_CMDLINE, _("Unreadable config file: %r"), options.configfile)
        config.read(files=files)
    except LinkCheckerError as msg:
        # config error
        print_usage(str(msg))
    drop_privileges()
    # set up config object using options
    setup_config(config, options)
    # now sanitize the configuration
    config.sanitize()

    log.debug(LOG_CMDLINE, "configuration: %s", pprint.pformat(sorted(config.items())))

    # prepare checking queue
    aggregate = get_aggregate(config)
    if options.trace:
        # enable thread tracing
        config["trace"] = True
        # start trace in mainthread
        from .. import trace

        trace.trace_filter([r"^linkcheck"])
        trace.trace_on()
    # add urls to queue
    if options.stdin:
        for url in read_stdin_urls():
            aggregate_url(aggregate, url)
    elif options.url:
        for url in options.url:
            aggregate_url(aggregate, stripurl(url))
    else:
        log.warn(LOG_CMDLINE, _("no files or URLs given"))
    # set up profiling
    do_profile = False
    if options.profile:
        if has_profile:
            if os.path.exists(_profile):
                print(
                    _(
                        "Overwrite profiling file %(file)r?\n"
                        "Press Ctrl-C to cancel, RETURN to continue."
                    )
                    % {"file": _profile}
                )
                try:
                    input()
                except KeyboardInterrupt:
                    print("", _("Canceled."), file=sys.stderr, sep="\n")
                    sys.exit(1)
            do_profile = True
        else:
            log.warn(
                LOG_CMDLINE,
                _(
                    "The `yappi' Python module is not installed,"
                    " therefore the --profile option is disabled."
                ),
            )

    # finally, start checking
    if do_profile:
        import yappi

        yappi.start()
        check_urls(aggregate)
        yappi.stop()
        yappi.get_func_stats().save(_profile)
    else:
        check_urls(aggregate)
    if config["debugmemory"]:
        from .. import memoryutil

        if has_meliae:
            log.info(LOG_CMDLINE, _("Dumping memory statistics..."))
            filename = memoryutil.write_memory_dump()
            message = _("The memory dump has been written to `%(filename)s'.")
            log.info(LOG_CMDLINE, message % dict(filename=filename))
        else:
            log.warn(LOG_CMDLINE, memoryutil.MemoryDebugMsg)

    stats = config["logger"].stats
    # on internal errors, exit with status 2
    if stats.internal_errors:
        sys.exit(2)
    # on errors or printed warnings, exit with status 1
    if stats.errors or (stats.warnings_printed and config["warnings"]):
        sys.exit(1)



# Copyright (C) 2011-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


def parse_bookmark_data(data):
    """Return iterator for bookmarks of the form (url, name, line number).
    Bookmarks are not sorted.
    """
    name = None
    lineno = 0
    for line in data.splitlines():
        lineno += 1
        line = line.strip()
        if line.startswith("NAME="):
            name = line[5:]
        elif line.startswith("URL="):
            url = line[4:]
            if url and name is not None:
                yield (url, name, lineno)
        else:
            name = None



#
# Copyright (c) 2012, Albert Zeyer, www.az2000.de
# All rights reserved.
# file created 2011-04-15


# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met: 
# 
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer. 
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution. 
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


# This is a simple replacement for the standard Python exception handler (sys.excepthook).
# In addition to what the standard handler does, it also prints all referenced variables
# (no matter if local, global or builtin) of the code line of each stack frame.
# See below for some examples and some example output.

# https://github.com/albertz/py_better_exchook

import sys
import os
import keyword

pykeywords = set(keyword.kwlist)


def parse_py_statement(line):
    state = 0
    curtoken = ""
    spaces = " \t\n"
    ops = ".,;:+-*/%&=|(){}[]^<>"
    i = 0
    def _escape_char(c):
        if c == "n": return "\n"
        elif c == "t": return "\t"
        else: return c
    while i < len(line):
        c = line[i]
        i += 1
        if state == 0:
            if c in spaces: pass
            elif c in ops: yield ("op", c)
            elif c == "#": state = 6
            elif c == "\"": state = 1
            elif c == "'": state = 2
            else:
                curtoken = c
                state = 3
        elif state == 1: # string via "
            if c == "\\": state = 4
            elif c == "\"":
                yield ("str", curtoken)
                curtoken = ""
                state = 0
            else: curtoken += c
        elif state == 2: # string via '
            if c == "\\": state = 5
            elif c == "'":
                yield ("str", curtoken)
                curtoken = ""
                state = 0
            else: curtoken += c
        elif state == 3: # identifier
            if c in spaces + ops + "#\"'":
                yield ("id", curtoken)
                curtoken = ""
                state = 0
                i -= 1
            else: curtoken += c
        elif state == 4: # escape in "
            curtoken += _escape_char(c)
            state = 1
        elif state == 5: # escape in '
            curtoken += _escape_char(c)
            state = 2
        elif state == 6: # comment
            curtoken += c
    if state == 3: yield ("id", curtoken)
    elif state == 6: yield ("comment", curtoken)


def grep_full_py_identifiers(tokens):
    global pykeywords
    tokens = list(tokens)
    i = 0
    while i < len(tokens):
        tokentype, token = tokens[i]
        i += 1
        if tokentype != "id": continue
        while i+1 < len(tokens) and tokens[i] == ("op", ".") and tokens[i+1][0] == "id":
            token += "." + tokens[i+1][1]
            i += 2
        if token == "": continue
        if token in pykeywords: continue
        if token[0] in ".0123456789": continue
        yield token

def output(s, out=sys.stdout): print(s, file=out)

def output_limit():
    return 300

def pp_extra_info(obj, depthlimit = 3):
    s = []
    if hasattr(obj, "__len__"):
        try:
            if type(obj) in (bytes,str,list,tuple,dict) and len(obj) <= 5:
                pass # don't print len in this case
            else:
                s += ["len = " + str(obj.__len__())]
        except Exception:
            pass
    if depthlimit > 0 and hasattr(obj, "__getitem__"):
        try:
            if type(obj) in (bytes,str):
                pass # doesn't make sense to get subitems here
            else:
                subobj = obj.__getitem__(0)
                extra_info = pp_extra_info(subobj, depthlimit - 1)
                if extra_info != "":
                    s += ["_[0]: {" + extra_info + "}"]
        except Exception:
            pass
    return ", ".join(s)
    
def pretty_print(obj):
    s = repr(obj)
    limit = output_limit()
    if len(s) > limit:
        s = s[:limit - 3] + "..."
    extra_info = pp_extra_info(obj)
    if extra_info != "": s += ", " + extra_info
    return s

def fallback_findfile(filename):
    mods = [ m for m in sys.modules.values() if m and hasattr(m, "__file__") and filename in m.__file__ ]
    if len(mods) == 0: return None
    altfn = mods[0].__file__
    if altfn[-4:-1] == ".py": altfn = altfn[:-1] # *.pyc or whatever
    return altfn

def better_exchook(etype, value, tb, out=sys.stdout):
    output('Traceback (most recent call last):', out=out)
    allLocals,allGlobals = {},{}
    try:
        import linecache
        limit = None
        if hasattr(sys, 'tracebacklimit'):
            limit = sys.tracebacklimit
        n = 0
        _tb = tb
        def _resolveIdentifier(namespace, id):
            obj = namespace[id[0]]
            for part in id[1:]:
                obj = getattr(obj, part)
            return obj
        def _trySet(old, prefix, func):
            if old is not None: return old
            try: return prefix + func()
            except KeyError: return old
            except Exception as e:
                return prefix + "!" + e.__class__.__name__ + ": " + str(e)
        while _tb is not None and (limit is None or n < limit):
            f = _tb.tb_frame
            allLocals.update(f.f_locals)
            allGlobals.update(f.f_globals)
            lineno = _tb.tb_lineno
            co = f.f_code
            filename = co.co_filename
            name = co.co_name
            output('  File "%s", line %d, in %s' % (filename,lineno,name), out=out)
            if not os.path.isfile(filename):
                altfn = fallback_findfile(filename)
                if altfn:
                    output("    -- couldn't find file, trying this instead: " + altfn, out=out)
                    filename = altfn
            linecache.checkcache(filename)
            line = linecache.getline(filename, lineno, f.f_globals)
            if line:
                line = line.strip()
                output('    line: ' + line, out=out)
                output('    locals:', out=out)
                alreadyPrintedLocals = set()
                for tokenstr in grep_full_py_identifiers(parse_py_statement(line)):
                    splittedtoken = tuple(tokenstr.split("."))
                    for token in map(lambda i: splittedtoken[0:i], range(1, len(splittedtoken) + 1)):
                        if token in alreadyPrintedLocals: continue
                        tokenvalue = None
                        tokenvalue = _trySet(tokenvalue, "<local> ", lambda: pretty_print(_resolveIdentifier(f.f_locals, token)))
                        tokenvalue = _trySet(tokenvalue, "<global> ", lambda: pretty_print(_resolveIdentifier(f.f_globals, token)))
                        tokenvalue = _trySet(tokenvalue, "<builtin> ", lambda: pretty_print(_resolveIdentifier(f.f_builtins, token)))
                        tokenvalue = tokenvalue or "<not found>"
                        output('      ' + ".".join(token) + " = " + tokenvalue, out=out)
                        alreadyPrintedLocals.add(token)
                if len(alreadyPrintedLocals) == 0: output("       no locals", out=out)
            else:
                output('    -- code not available --', out=out)
            _tb = _tb.tb_next
            n += 1

    except Exception:
        output("ERROR: cannot get more detailed exception info because:", out=out)
        import traceback
        for l in traceback.format_exc().split("\n"): output("   " + l, out=out)
        output("simple traceback:", out=out)
        traceback.print_tb(tb, None, out)

    import types
    def _some_str(value):
        try:
            return str(value)
        except Exception:
            return '<unprintable %s object>' % type(value).__name__

    def _format_final_exc_line(etype, value):
        valuestr = _some_str(value)
        if value is None or not valuestr:
            line = "%s" % etype
        else:
            line = f"{etype}: {valuestr}"
        return line
    if (isinstance(etype, BaseException) or
        (hasattr(types, "InstanceType") and isinstance(etype, types.InstanceType)) or
        etype is None or type(etype) is str):
        output(_format_final_exc_line(etype, value), out=out)
    else:
        output(_format_final_exc_line(etype.__name__, value), out=out)

def install():
    sys.excepthook = better_exchook



# Copyright (C) 2010-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Parse hyperlinks in Word files.
"""
from . import _ParserPlugin

try:
    import win32com
    import pythoncom

    has_win32com = True
    Error = pythoncom.com_error
except ImportError:
    has_win32com = False
    Error = Exception
from .. import fileutil, log, LOG_PLUGIN


_initialized = False


def init_win32com():
    """Initialize the win32com.client cache."""
    global _initialized
    if _initialized:
        return
    import win32com.client

    if win32com.client.gencache.is_readonly:
        # allow gencache to create the cached wrapper objects
        win32com.client.gencache.is_readonly = False
        # under py2exe the call in gencache to __init__() does not happen
        # so we use Rebuild() to force the creation of the gen_py folder
        # Note that the python...\win32com.client.gen_py dir must not exist
        # to allow creation of the cache in %temp% for py2exe.
        # This is ensured by excluding win32com.gen_py in setup.py
        win32com.client.gencache.Rebuild()
    _initialized = True


def has_word():
    """Determine if Word is available on the current system."""
    if not has_win32com:
        return False
    try:
        import _winreg as winreg
    except ImportError:
        import winreg
    try:
        key = winreg.OpenKey(winreg.HKEY_CLASSES_ROOT, "Word.Application")
        winreg.CloseKey(key)
        return True
    except (OSError, ImportError):
        pass
    return False


def constants(name):
    """Helper to return constants. Avoids importing win32com.client in
    other modules."""
    return getattr(win32com.client.constants, name)


def get_word_app():
    """Return open Word.Application handle, or None if Word is not available
    on this system."""
    if not has_word():
        return None
    # Since this function is called from different threads, initialize
    # the COM layer.
    pythoncom.CoInitialize()
    import win32com.client

    app = win32com.client.gencache.EnsureDispatch("Word.Application")
    app.Visible = False
    return app


def close_word_app(app):
    """Close Word application object."""
    app.Quit()


def open_wordfile(app, filename):
    """Open given Word file with application object."""
    return app.Documents.Open(
        filename,
        ReadOnly=True,
        AddToRecentFiles=False,
        Visible=False,
        NoEncodingDialog=True,
    )


def close_wordfile(doc):
    """Close word file."""
    doc.Close()


class WordParser(_ParserPlugin):
    """Word parsing plugin."""

    def __init__(self, config):
        """Check for Word."""
        init_win32com()
        if not has_word():
            log.warn(LOG_PLUGIN, "Microsoft Word not found for WordParser plugin")
        super().__init__(config)

    def applies_to(self, url_data, pagetype=None):
        """Check for Word pagetype."""
        return has_word() and pagetype == 'word'

    def check(self, url_data):
        """Parse Word data."""
        content = url_data.get_raw_content()
        filename = get_temp_filename(content)
        # open word file and parse hyperlinks
        try:
            app = get_word_app()
            try:
                doc = open_wordfile(app, filename)
                if doc is None:
                    raise Error("could not open word file %r" % filename)
                try:
                    for link in doc.Hyperlinks:
                        line = get_line_number(link.Range)
                        name = link.TextToDisplay
                        url_data.add_url(link.Address, name=name, line=line)
                finally:
                    close_wordfile(doc)
            finally:
                close_word_app(app)
        except Error as msg:
            log.warn(LOG_PLUGIN, "Error parsing word file: %s", msg)


def get_line_number(doc, wrange):
    """Get line number for given range object."""
    lineno = 1
    wrange.Select()
    wdFirstCharacterLineNumber = constants("wdFirstCharacterLineNumber")
    wdGoToLine = constants("wdGoToLine")
    wdGoToPrevious = constants("wdGoToPrevious")
    while True:
        curline = doc.Selection.Information(wdFirstCharacterLineNumber)
        doc.Selection.GoTo(wdGoToLine, wdGoToPrevious, Count=1, Name="")
        lineno += 1
        prevline = doc.Selection.Information(wdFirstCharacterLineNumber)
        if prevline == curline:
            break
    return lineno


def get_temp_filename(content):
    """Get temporary filename for content to parse."""
    # store content in temporary file
    fd, filename = fileutil.get_temp_file(mode='wb', suffix='.doc', prefix='lc_')
    try:
        fd.write(content)
    finally:
        fd.close()
    return filename



# Copyright (C) 2000-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Main package for link checking.
"""

# version checks
import sys

if sys.version_info < (3, 7, 0, 'final', 0):
    import platform

    raise SystemExit(
        "This program requires Python 3.7 or later instead of %s."
        % platform.python_version()
    )

import os
import re

from . import i18n, log
from .logconf import (
    LOG_ROOT,
    LOG_CMDLINE,
    LOG_CHECK,
    LOG_CACHE,
    LOG_THREAD,
    LOG_PLUGIN,
)

COMMAND_NAME = "linkchecker"
PACKAGE_NAME = __package__


def module_path():
    """Return absolute directory of system executable."""
    return os.path.dirname(os.path.abspath(sys.executable))


class LinkCheckerError(Exception):
    """Exception to be raised on linkchecker-specific check errors."""

    pass


class LinkCheckerInterrupt(Exception):
    """Used for testing."""

    pass


def get_link_pat(arg, strict=False):
    """Get a link pattern matcher for intern/extern links.
    Returns a compiled pattern and a negate and strict option.

    @param arg: pattern from config
    @type arg: string
    @param strict: if pattern is to be handled strict
    @type strict: bool
    @return: dictionary with keys 'pattern', 'negate' and 'strict'
    @rtype: dict
    @raises: re.error on invalid regular expressions
    """
    log.debug(LOG_CHECK, _("Link pattern %r strict=%s"), arg, strict)
    if arg.startswith('!'):
        pattern = arg[1:]
        negate = True
    else:
        pattern = arg
        negate = False
    try:
        regex = re.compile(pattern)
    except re.error as msg:
        log.warn(LOG_CHECK, _("invalid regular expression %r: %s"), pattern, msg)
        raise
    return {
        "pattern": regex,
        "negate": negate,
        "strict": strict,
    }


def init_i18n(loc=None):
    """Initialize i18n with the configured locale dir. The environment
    variable LOCPATH can also specify a locale dir.

    @return: None
    """
    if 'LOCPATH' in os.environ:
        locdir = os.environ['LOCPATH']
    else:
        # Need Python 3.9 for importlib.resources.files
        locdir = os.path.join(__path__[0], 'data', 'locale')
    i18n.init(COMMAND_NAME, locdir, loc=loc)
    # install translated log level names
    import logging

    logging.addLevelName(logging.CRITICAL, _('CRITICAL'))
    logging.addLevelName(logging.ERROR, _('ERROR'))
    logging.addLevelName(logging.WARN, _('WARN'))
    logging.addLevelName(logging.WARNING, _('WARNING'))
    logging.addLevelName(logging.INFO, _('INFO'))
    logging.addLevelName(logging.DEBUG, _('DEBUG'))
    logging.addLevelName(logging.NOTSET, _('NOTSET'))


# initialize i18n, puts _() and _n() function into global namespace
init_i18n()



# Copyright (C) 2012-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Memory utilities.
"""
import gc
import pprint
from . import strformat, log, LOG_CHECK
from .fileutil import get_temp_file

# Message to display when meliae package is not installed
MemoryDebugMsg = strformat.format_feature_warning(
    module='meliae', feature='memory debugging', url='https://launchpad.net/meliae'
)


def write_memory_dump():
    """Dump memory to a temporary filename with the meliae package.

    @return: JSON filename where memory dump has been written to
    @rtype: string
    """
    # first do a full garbage collection run
    gc.collect()
    if gc.garbage:
        log.warn(LOG_CHECK, "Unreachabe objects: %s", pprint.pformat(gc.garbage))
    from meliae import scanner

    fo, filename = get_temp_file(mode='wb', suffix='.json', prefix='lcdump_')
    try:
        scanner.dump_all_objects(fo)
    finally:
        fo.close()
    return filename



# Copyright (C) 2000-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Application internationalization support.
"""

# i18n support
import os
import locale
import gettext
import sys
import codecs

# more supported languages are added in init()
supported_languages = {'en'}
default_language = default_encoding = None
default_directory = None
default_domain = None


def install_builtin(translator, do_unicode):
    """Install _() and _n() gettext methods into default namespace."""
    import builtins

    builtins.__dict__['_'] = translator.gettext
    # also install ngettext
    builtins.__dict__['_n'] = translator.ngettext


class Translator(gettext.GNUTranslations):
    """A translation class always installing its gettext methods into the
    default namespace."""

    def install(self, do_unicode):
        """Install gettext methods into the default namespace."""
        install_builtin(self, do_unicode)


class NullTranslator(gettext.NullTranslations):
    """A dummy translation class always installing its gettext methods into
    the default namespace."""

    def install(self, do_unicode):
        """Install gettext methods into the default namespace."""
        install_builtin(self, do_unicode)


def init(domain, directory, loc=None):
    """Initialize this gettext i18n module. Searches for supported languages
    and installs the gettext translator class."""
    global default_language, default_encoding, default_domain, default_directory
    default_directory = directory
    default_domain = domain
    if os.path.isdir(directory):
        # get supported languages
        for lang in os.listdir(directory):
            path = os.path.join(directory, lang, 'LC_MESSAGES')
            mo_file = os.path.join(path, f"{domain}.mo")
            if os.path.exists(mo_file):
                supported_languages.add(lang)
    if loc is None:
        loc, encoding = get_locale()
    else:
        encoding = get_locale()[1]
    if loc in supported_languages:
        default_language = loc
    else:
        default_language = "en"
    # Even if the default language is not supported, the encoding should
    # be installed. Otherwise the Python installation is borked.
    default_encoding = encoding
    install_language(default_language)


def install_language(language):
    """Install translation service routines into default namespace."""
    translator = get_translator(
        default_domain, default_directory, languages=[get_lang(language)], fallback=True
    )
    do_unicode = True
    translator.install(do_unicode)


def get_translator(
    domain,
    directory,
    languages=None,
    translatorklass=Translator,
    fallback=False,
    fallbackklass=NullTranslator,
):
    """Search the appropriate GNUTranslations class."""
    translator = gettext.translation(
        domain,
        localedir=directory,
        languages=languages,
        class_=translatorklass,
        fallback=fallback,
    )
    if not isinstance(translator, gettext.GNUTranslations) and fallbackklass:
        translator = fallbackklass()
    return translator


def get_lang(lang):
    """Return lang if it is supported, or the default language."""
    if lang in supported_languages:
        return lang
    return default_language


def get_locale():
    """Search the default platform locale and norm it.
    @returns (locale, encoding)
    @rtype (string, string)"""
    try:
        loc, encoding = locale.getdefaultlocale()
    except ValueError:
        # locale configuration is broken - ignore that
        loc, encoding = None, None
    if loc is None:
        loc = "C"
    else:
        loc = norm_locale(loc)
    if encoding is None:
        encoding = "ascii"
    return (loc, encoding)


def norm_locale(loc):
    """Normalize a locale."""
    loc = locale.normalize(loc)
    # split up the locale into its base components
    pos = loc.find('@')
    if pos >= 0:
        loc = loc[:pos]
    pos = loc.find('.')
    if pos >= 0:
        loc = loc[:pos]
    pos = loc.find('_')
    if pos >= 0:
        loc = loc[:pos]
    return loc


def get_encoded_writer(out=sys.stdout, encoding=None, errors='replace'):
    """Get wrapped output writer with given encoding and error handling."""
    if encoding is None:
        encoding = default_encoding
    Writer = codecs.getwriter(encoding)
    return Writer(out.buffer, errors)



# Copyright (C) 2000-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
A HTML logger.
"""

import html
import os
import time

from . import _Logger
from .. import strformat, configuration


# ss=1 enables show source
validate_html = "https://validator.w3.org/check?ss=1&amp;uri=%(uri)s"
# options are the default
validate_css = (
    "https://jigsaw.w3.org/css-validator/validator?"
    "uri=%(uri)s&amp;warning=1&amp;profile=css2&amp;usermedium=all"
)

HTML_HEADER = """<!DOCTYPE HTML>
<html>
<head>
<meta http-equiv="Content-Type" content="text/html; charset=%(encoding)s"/>
<title>%(title)s</title>
<style type="text/css">
<!--
 h2 { font-family: Verdana,sans-serif; font-size: 22pt; font-weight: bold; }
 body { font-family: Arial,sans-serif; font-size: 11pt; background-color: %(body)s; }
 td { font-family: Arial,sans-serif; font-size: 11pt; }
 code { font-family: Courier; }
 a:link {color: %(link)s;}
 a:visited {color: %(vlink)s;}
 a:active {color: %(alink)s;}
 a:hover { color: #34a4ef; }
 table { border-collapse:collapse; }
 table, th, td { border: 1px solid black; padding: 2px; }
 td.url { background-color: %(url)s }
 td.valid { background-color: %(valid)s }
 td.error { background-color: %(error)s }
 td.warning { background-color: %(warning)s }
-->
</style>
</head>
<body>
"""


class HtmlLogger(_Logger):
    """Logger with HTML output."""

    LoggerName = 'html'

    LoggerArgs = {
        "filename": "linkchecker-out.html",
        'colorbackground': '#fff7e5',
        'colorurl': '#dcd5cf',
        'colorborder': '#000000',
        'colorlink': '#191c83',
        'colorwarning': '#e0954e',
        'colorerror': '#db4930',
        'colorok': '#3ba557',
    }

    def __init__(self, **kwargs):
        """Initialize default HTML color values."""
        args = self.get_args(kwargs)
        super().__init__(**args)
        self.init_fileoutput(args)
        self.colorbackground = args['colorbackground']
        self.colorurl = args['colorurl']
        self.colorborder = args['colorborder']
        self.colorlink = args['colorlink']
        self.colorwarning = args['colorwarning']
        self.colorerror = args['colorerror']
        self.colorok = args['colorok']

    def part(self, name):
        """Return non-space-breakable part name."""
        return super().part(name).replace(" ", "&nbsp;")

    def comment(self, s, **args):
        """Write HTML comment."""
        self.write("<!-- ")
        self.write(s, **args)
        self.write(" -->")

    def start_output(self):
        """Write start of checking info."""
        super().start_output()
        header = {
            "encoding": self.get_charset_encoding(),
            "title": configuration.App,
            "body": self.colorbackground,
            "link": self.colorlink,
            "vlink": self.colorlink,
            "alink": self.colorlink,
            "url": self.colorurl,
            "error": self.colorerror,
            "valid": self.colorok,
            "warning": self.colorwarning,
        }
        self.write(HTML_HEADER % header)
        self.comment("Generated by %s" % configuration.App)
        if self.has_part('intro'):
            self.write(
                "<h2>"
                + configuration.App
                + "</h2><br/><blockquote>"
                + configuration.Freeware
                + "<br/><br/>"
                + (_("Start checking at %s") % strformat.strtime(self.starttime))
                + os.linesep
                + "<br/>"
            )
            self.check_date()
        self.flush()

    def log_url(self, url_data):
        """Write url checking info as HTML."""
        self.write_table_start()
        if self.has_part("url"):
            self.write_url(url_data)
        if url_data.name and self.has_part("name"):
            self.write_name(url_data)
        if url_data.parent_url and self.has_part("parenturl"):
            self.write_parent(url_data)
        if url_data.base_ref and self.has_part("base"):
            self.write_base(url_data)
        if url_data.url and self.has_part("realurl"):
            self.write_real(url_data)
        if url_data.dltime >= 0 and self.has_part("dltime"):
            self.write_dltime(url_data)
        if url_data.size >= 0 and self.has_part("dlsize"):
            self.write_size(url_data)
        if url_data.checktime and self.has_part("checktime"):
            self.write_checktime(url_data)
        if url_data.info and self.has_part("info"):
            self.write_info(url_data)
        if url_data.modified and self.has_part("modified"):
            self.write_modified(url_data)
        if url_data.warnings and self.has_part("warning"):
            self.write_warning(url_data)
        if self.has_part("result"):
            self.write_result(url_data)
        self.write_table_end()
        self.flush()

    def write_table_start(self):
        """Start html table."""
        self.writeln('<br/><br/><table>')

    def write_table_end(self):
        """End html table."""
        self.write('</table><br/>')

    def write_id(self):
        """Write ID for current URL."""
        self.writeln("<tr>")
        self.writeln('<td>%s</td>' % self.part("id"))
        self.write("<td>%d</td></tr>" % self.stats.number)

    def write_url(self, url_data):
        """Write url_data.base_url."""
        self.writeln("<tr>")
        self.writeln('<td class="url">%s</td>' % self.part("url"))
        self.write('<td class="url">')
        self.write("`%s'" % html.escape(url_data.base_url))
        self.writeln("</td></tr>")

    def write_name(self, url_data):
        """Write url_data.name."""
        args = (self.part("name"), html.escape(url_data.name))
        self.writeln("<tr><td>%s</td><td>`%s'</td></tr>" % args)

    def write_parent(self, url_data):
        """Write url_data.parent_url."""
        self.write(
            "<tr><td>"
            + self.part("parenturl")
            + '</td><td><a target="top" href="'
            + html.escape(url_data.parent_url)
            + '">'
            + html.escape(url_data.parent_url)
            + "</a>"
        )
        if url_data.line is not None:
            self.write(_(", line %d") % url_data.line)
        if url_data.column is not None:
            self.write(_(", col %d") % url_data.column)
        if url_data.page > 0:
            self.write(_(", page %d") % url_data.page)
        if not url_data.valid:
            # on errors show HTML and CSS validation for parent url
            vhtml = validate_html % {'uri': url_data.parent_url}
            vcss = validate_css % {'uri': url_data.parent_url}
            self.writeln()
            self.writeln('(<a href="' + vhtml + '">HTML</a>)')
            self.write('(<a href="' + vcss + '">CSS</a>)')
        self.writeln("</td></tr>")

    def write_base(self, url_data):
        """Write url_data.base_ref."""
        self.writeln(
            "<tr><td>"
            + self.part("base")
            + "</td><td>"
            + html.escape(url_data.base_ref)
            + "</td></tr>"
        )

    def write_real(self, url_data):
        """Write url_data.url."""
        self.writeln(
            "<tr><td>"
            + self.part("realurl")
            + "</td><td>"
            + '<a target="top" href="'
            + html.escape(url_data.url)
            + '">'
            + html.escape(url_data.url)
            + "</a></td></tr>"
        )

    def write_dltime(self, url_data):
        """Write url_data.dltime."""
        self.writeln(
            "<tr><td>"
            + self.part("dltime")
            + "</td><td>"
            + (_("%.3f seconds") % url_data.dltime)
            + "</td></tr>"
        )

    def write_size(self, url_data):
        """Write url_data.size."""
        self.writeln(
            "<tr><td>"
            + self.part("dlsize")
            + "</td><td>"
            + strformat.strsize(url_data.size)
            + "</td></tr>"
        )

    def write_checktime(self, url_data):
        """Write url_data.checktime."""
        self.writeln(
            "<tr><td>"
            + self.part("checktime")
            + "</td><td>"
            + (_("%.3f seconds") % url_data.checktime)
            + "</td></tr>"
        )

    def write_info(self, url_data):
        """Write url_data.info."""
        sep = "<br/>" + os.linesep
        text = sep.join(html.escape(x) for x in url_data.info)
        self.writeln(
            '<tr><td valign="top">'
            + self.part("info")
            + "</td><td>"
            + text
            + "</td></tr>"
        )

    def write_modified(self, url_data):
        """Write url_data.modified."""
        text = html.escape(self.format_modified(url_data.modified))
        self.writeln(
            '<tr><td valign="top">'
            + self.part("modified")
            + "</td><td>"
            + text
            + "</td></tr>"
        )

    def write_warning(self, url_data):
        """Write url_data.warnings."""
        sep = "<br/>" + os.linesep
        text = sep.join(html.escape(x[1]) for x in url_data.warnings)
        self.writeln(
            '<tr><td class="warning" valign="top">'
            + self.part("warning")
            + '</td><td class="warning">'
            + text
            + "</td></tr>"
        )

    def write_result(self, url_data):
        """Write url_data.result."""
        if url_data.valid:
            self.write('<tr><td class="valid">')
            self.write(self.part("result"))
            self.write('</td><td class="valid">')
            self.write(html.escape(_("Valid")))
        else:
            self.write('<tr><td class="error">')
            self.write(self.part("result"))
            self.write('</td><td class="error">')
            self.write(html.escape(_("Error")))
        if url_data.result:
            self.write(": " + html.escape(url_data.result))
        self.writeln("</td></tr>")

    def write_stats(self):
        """Write check statistic infos."""
        self.writeln('<br/><i>%s</i><br/>' % _("Statistics"))
        if self.stats.number > 0:
            self.writeln(
                _(
                    "Content types: %(image)d image, %(text)d text, %(video)d video, "
                    "%(audio)d audio, %(application)d application, %(mail)d mail"
                    " and %(other)d other."
                )
                % self.stats.link_types
            )
            self.writeln("<br/>")
            self.writeln(
                _("URL lengths: min=%(min)d, max=%(max)d, avg=%(avg)d.")
                % dict(
                    min=self.stats.min_url_length,
                    max=self.stats.max_url_length,
                    avg=self.stats.avg_url_length,
                )
            )
        else:
            self.writeln(_("No statistics available since no URLs were checked."))
        self.writeln("<br/>")

    def write_outro(self):
        """Write end of check message."""
        self.writeln("<br/>")
        self.write(_("That's it.") + " ")
        if self.stats.number >= 0:
            self.write(
                _n("%d link checked.", "%d links checked.", self.stats.number)
                % self.stats.number
            )
            self.write(" ")
        self.write(
            _n("%d warning found", "%d warnings found", self.stats.warnings_printed)
            % self.stats.warnings_printed
        )
        if self.stats.warnings != self.stats.warnings_printed:
            self.write(
                _(" (%d ignored or duplicates not printed)")
                % (self.stats.warnings - self.stats.warnings_printed)
            )
        self.write(". ")
        self.write(
            _n("%d error found", "%d errors found", self.stats.errors_printed)
            % self.stats.errors_printed
        )
        if self.stats.errors != self.stats.errors_printed:
            self.write(
                _(" (%d duplicates not printed)")
                % (self.stats.errors - self.stats.errors_printed)
            )
        self.writeln(".")
        self.writeln("<br/>")
        num = self.stats.internal_errors
        if num:
            self.write(
                _n(
                    "There was %(num)d internal error.",
                    "There were %(num)d internal errors.",
                    num,
                )
                % {"num": num}
            )
            self.writeln("<br/>")
        self.stoptime = time.time()
        duration = self.stoptime - self.starttime
        self.writeln(
            _("Stopped checking at %(time)s (%(duration)s)")
            % {
                "time": strformat.strtime(self.stoptime),
                "duration": strformat.strduration_long(duration),
            }
        )
        self.writeln(
            '</blockquote><br/><hr><small>' + configuration.HtmlAppInfo + "<br/>"
        )
        self.writeln(
            _("Read the documentation at %s")
            % (
                '<a href="'
                + configuration.Url
                + '" target="_top">'
                + configuration.Url
                + "</a>.<br/>"
            )
        )
        self.writeln(
            _("Write comments and bugs to %s")
            % (
                '<a href="'
                + configuration.SupportUrl
                + '">'
                + configuration.SupportUrl
                + "</a>.<br/>"
            )
        )
        self.writeln("</small></body></html>")

    def end_output(self, **kwargs):
        """Write end of checking info as HTML."""
        if self.has_part("stats"):
            self.write_stats()
        if self.has_part("outro"):
            self.write_outro()
        self.close_fileoutput()



# Copyright (C) 2000-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Utility functions suitable for command line clients.
"""
import sys
import argparse
from . import checker, fileutil, strformat, plugins
from .director import console


class LCArgumentParser(argparse.ArgumentParser):
    """Custom argument parser to format help text."""

    def print_help(self, file=sys.stdout):
        """Print a help message to stdout."""
        msg = self.format_help()
        if fileutil.is_tty(file):
            strformat.paginate(msg)
        else:
            print(msg, file=file)


def print_version(exit_code=0):
    """Print the program version and exit."""
    console.print_version()
    sys.exit(exit_code)


def print_plugins(folders, exit_code=0):
    """Print available plugins and exit."""
    modules = plugins.get_plugin_modules(folders)
    pluginclasses = sorted(
        plugins.get_plugin_classes(modules), key=lambda x: x.__name__
    )

    for pluginclass in pluginclasses:
        print(pluginclass.__name__)
        doc = strformat.wrap(pluginclass.__doc__, 80)
        print(strformat.indent(doc))
        print()
    sys.exit(exit_code)


def print_usage(msg, exit_code=2):
    """Print a program msg text to stderr and exit."""
    program = sys.argv[0]
    print(_("Error: %(msg)s") % {"msg": msg}, file=console.stderr)
    print(
        _("Execute '%(program)s -h' for help") % {"program": program},
        file=console.stderr,
    )
    sys.exit(exit_code)


def aggregate_url(aggregate, url, err_exit_code=2):
    """Append given commandline URL to input queue."""
    get_url_from = checker.get_url_from
    url = checker.guess_url(url)
    url_data = get_url_from(url, 0, aggregate, extern=(0, 0))
    aggregate.urlqueue.put(url_data)



# Copyright (C) 2005-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Locking utility class.
"""
import threading
from . import log, LOG_THREAD


def get_lock(name, debug=False):
    """Get a new lock.

    @param debug: if True, acquire() and release() will have debug messages
    @type debug: boolean, default is False
    @return: a lock object
    @rtype: threading.Lock or DebugLock
    """
    lock = threading.Lock()
    # for thread debugging, use the DebugLock wrapper
    if debug:
        lock = DebugLock(lock, name)
    return lock


class DebugLock:
    """Debugging lock class."""

    def __init__(self, lock, name):
        """Store lock and name parameters."""
        self.lock = lock
        self.name = name

    def acquire(self, blocking=1):
        """Acquire lock."""
        threadname = threading.currentThread().name
        log.debug(LOG_THREAD, "Acquire %s for %s", self.name, threadname)
        self.lock.acquire(blocking)
        log.debug(LOG_THREAD, "...acquired %s for %s", self.name, threadname)

    def release(self):
        """Release lock."""
        threadname = threading.currentThread().name
        log.debug(LOG_THREAD, "Release %s for %s", self.name, threadname)
        self.lock.release()


def get_semaphore(name, value=None, debug=False):
    """Get a new semaphore.

    @param value: if not None, a BoundedSemaphore will be used
    @type debug: int or None
    @param debug: if True, acquire() and release() will have debug messages
    @type debug: boolean, default is False
    @return: a semaphore object
    @rtype: threading.Semaphore or threading.BoundedSemaphore or DebugLock
    """
    if value is None:
        lock = threading.Semaphore()
    else:
        lock = threading.BoundedSemaphore(value)
    if debug:
        lock = DebugLock(lock, name)
    return lock



# Copyright (C) 2000-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Handle https links.
"""
import time
import threading
from . import _ConnectionPlugin
from .. import strformat, LinkCheckerError
from ..decorators import synchronized

_lock = threading.Lock()

# configuration option names
sslcertwarndays = "sslcertwarndays"


class SslCertificateCheck(_ConnectionPlugin):
    """Check SSL certificate expiration date. Only internal https: links
    will be checked. A domain will only be checked once to avoid duplicate
    warnings.
    The expiration warning time can be configured with the sslcertwarndays
    option."""

    def __init__(self, config):
        """Initialize clamav configuration."""
        super().__init__(config)
        self.warn_ssl_cert_secs_valid = (
            config[sslcertwarndays] * strformat.SECONDS_PER_DAY
        )
        # do not check hosts multiple times
        self.checked_hosts = set()

    def applies_to(self, url_data):
        """Check validity, scheme, extern and url_connection."""
        return (
            url_data.valid
            and url_data.scheme == 'https'
            and not url_data.extern[0]
            and url_data.url_connection is not None
        )

    @synchronized(_lock)
    def check(self, url_data):
        """Run all SSL certificate checks that have not yet been done.
        OpenSSL already checked the SSL notBefore and notAfter dates.
        """
        host = url_data.urlparts[1]
        if host in self.checked_hosts:
            return
        self.checked_hosts.add(host)
        cert = url_data.ssl_cert
        config = url_data.aggregate.config
        if cert and 'notAfter' in cert:
            self.check_ssl_valid_date(url_data, cert)
        elif config['sslverify']:
            msg = _('certificate did not include "notAfter" information')
            url_data.add_warning(msg)
        else:
            msg = _('SSL verification is disabled; enable the sslverify option')
            url_data.add_warning(msg)

    def check_ssl_valid_date(self, url_data, cert):
        """Check if the certificate is still valid, or if configured check
        if it's at least a number of days valid.
        """
        import ssl

        try:
            notAfter = ssl.cert_time_to_seconds(cert['notAfter'])
        except ValueError as msg:
            msg = _('Invalid SSL certificate "notAfter" value %r') % cert['notAfter']
            url_data.add_warning(msg)
            return
        curTime = time.time()
        # Calculate seconds until certificate expires. Can be negative if
        # the certificate is already expired.
        secondsValid = notAfter - curTime
        args = dict(expire=cert['notAfter'])
        if secondsValid < 0:
            msg = _('SSL certificate is expired on %(expire)s.')
            url_data.add_warning(msg % args)
        else:
            args['valid'] = strformat.strduration_long(secondsValid)
            if secondsValid < self.warn_ssl_cert_secs_valid:
                msg = _(
                    'SSL certificate expires on %(expire)s and is only %(valid)s valid.'
                )
                url_data.add_warning(msg % args)
            else:
                msg = _('SSL certificate expires on %(expire)s and is %(valid)s valid.')
                url_data.add_info(msg % args)

    @classmethod
    def read_config(cls, configparser):
        """Read configuration file options."""
        config = dict()
        section = cls.__name__
        option = sslcertwarndays
        if configparser.has_option(section, option):
            num = configparser.getint(section, option)
            if num > 0:
                config[option] = num
            else:
                msg = _("invalid value for %s: %d must not be less than %d") % (
                    option,
                    num,
                    0,
                )
                raise LinkCheckerError(msg)
        else:
            # set the default
            config[option] = 30
        return config



# Copyright (C) 2000-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
An XML logger.
"""
from . import xmllog
from .. import strformat


class CustomXMLLogger(xmllog._XMLLogger):
    """
    XML custom output for easy post-processing.
    """

    LoggerName = "xml"

    LoggerArgs = {
        "filename": "linkchecker-out.xml",
    }

    def start_output(self):
        """
        Write start of checking info as xml comment.
        """
        super().start_output()
        self.xml_start_output()
        attrs = {"created": strformat.strtime(self.starttime)}
        self.xml_starttag('linkchecker', attrs)
        self.flush()

    def log_url(self, url_data):
        """
        Log URL data in custom XML format.
        """
        self.xml_starttag('urldata')
        if self.has_part('url'):
            self.xml_tag("url", url_data.base_url)
        if url_data.name and self.has_part('name'):
            self.xml_tag("name", url_data.name)
        if url_data.parent_url and self.has_part('parenturl'):
            attrs = {
                'line': "%s" % url_data.line,
                'column': "%s" % url_data.column,
            }
            self.xml_tag("parent", url_data.parent_url, attrs=attrs)
        if url_data.base_ref and self.has_part('base'):
            self.xml_tag("baseref", url_data.base_ref)
        if self.has_part("realurl"):
            self.xml_tag("realurl", url_data.url)
        if self.has_part("extern"):
            self.xml_tag("extern", "%d" % (1 if url_data.extern else 0))
        if url_data.dltime >= 0 and self.has_part("dltime"):
            self.xml_tag("dltime", "%f" % url_data.dltime)
        if url_data.size >= 0 and self.has_part("dlsize"):
            self.xml_tag("dlsize", "%d" % url_data.size)
        if url_data.checktime and self.has_part("checktime"):
            self.xml_tag("checktime", "%f" % url_data.checktime)
        if self.has_part("level"):
            self.xml_tag("level", "%d" % url_data.level)
        if url_data.info and self.has_part('info'):
            self.xml_starttag("infos")
            for info in url_data.info:
                self.xml_tag("info", info)
            self.xml_endtag("infos")
        if url_data.modified and self.has_part('modified'):
            self.xml_tag("modified", self.format_modified(url_data.modified))
        if url_data.warnings and self.has_part('warning'):
            self.xml_starttag("warnings")
            for tag, data in url_data.warnings:
                attrs = {}
                if tag:
                    attrs["tag"] = tag
                self.xml_tag("warning", data, attrs)
            self.xml_endtag("warnings")
        if self.has_part("result"):
            attrs = {}
            if url_data.result:
                attrs["result"] = url_data.result
            self.xml_tag("valid", "%d" % (1 if url_data.valid else 0), attrs)
        self.xml_endtag('urldata')
        self.flush()

    def end_output(self, **kwargs):
        """
        Write XML end tag.
        """
        self.xml_endtag("linkchecker")
        self.xml_end_output()
        self.close_fileoutput()



# Copyright (C) 2000-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""Parse configuration files"""

from configparser import RawConfigParser
from re import compile as re_compile
import os

from .. import (
    LinkCheckerError,
    get_link_pat,
    LOG_CHECK,
    log,
    fileutil,
    plugins,
    logconf,
)


def read_multiline(value):
    """Helper function reading multiline values."""
    for line in value.splitlines():
        line = line.strip()
        if not line or line.startswith('#'):
            continue
        yield line


class LCConfigParser(RawConfigParser):
    """
    Parse a LinkChecker configuration file.
    """

    def __init__(self, config):
        """Initialize configuration."""
        super().__init__()
        self.config = config

    def read(self, files):
        """Read settings from given config files.

        @raises: LinkCheckerError on syntax errors in the config file(s)
        """
        assert isinstance(files, list), "Invalid file list %r" % files
        try:
            self.read_ok = super().read(files)
            if len(self.read_ok) < len(files):
                failed_files = set(files) - set(self.read_ok)
                log.warn(
                    LOG_CHECK, "Could not read configuration files %s.", failed_files
                )
            # Read all the configuration parameters from the given files.
            self.read_checking_config()
            self.read_authentication_config()
            self.read_filtering_config()
            self.read_output_config()
            self.read_plugin_config()
        except Exception as msg:
            raise LinkCheckerError(_("Error parsing configuration: %s") % str(msg))

    def read_string_option(self, section, option, allowempty=False):
        """Read a string option."""
        if self.has_option(section, option):
            value = self.get(section, option)
            if not allowempty and not value:
                raise LinkCheckerError(
                    _("invalid empty value for %s: %s\n") % (option, value)
                )
            self.config[option] = value

    def read_boolean_option(self, section, option):
        """Read a boolean option."""
        if self.has_option(section, option):
            self.config[option] = self.getboolean(section, option)

    def read_float_option(self, section, option, key=None, min=None, max=None):
        """Read a float option."""
        if self.has_option(section, option):
            num = self.getfloat(section, option)
            if min is not None and num < min:
                raise LinkCheckerError(
                    _("invalid value for %s: %d must not be less than %d")
                    % (option, num, min)
                )
            if max is not None and num < max:
                raise LinkCheckerError(
                    _("invalid value for %s: %d must not be greater than %d")
                    % (option, num, max)
                )
            if key is None:
                key = option
            self.config[key] = num

    def read_int_option(self, section, option, key=None, min=None, max=None):
        """Read an integer option."""
        if self.has_option(section, option):
            num = self.getint(section, option)
            if min is not None and num < min:
                raise LinkCheckerError(
                    _("invalid value for %s: %d must not be less than %d")
                    % (option, num, min)
                )
            if max is not None and num < max:
                raise LinkCheckerError(
                    _("invalid value for %s: %d must not be greater than %d")
                    % (option, num, max)
                )
            if key is None:
                key = option
            self.config[key] = num

    def read_output_config(self):
        """Read configuration options in section "output"."""
        section = "output"
        from ..logger import LoggerClasses

        if self.has_section("blacklist"):
            log.warn(
                     LOG_CHECK,
                     _("The blacklist section in linkcheckerrc is deprecated, "
                       "please rename to failures")
            )
            for opt in self.options("blacklist"):
                self.config["failures"][opt] = self.get("blacklist", opt)
        for c in LoggerClasses:
            key = c.LoggerName
            if self.has_section(key):
                for opt in self.options(key):
                    self.config[key][opt] = self.get(key, opt)
                if self.has_option(key, 'parts'):
                    val = self.get(key, 'parts')
                    parts = [f.strip().lower() for f in val.split(',')]
                    self.config[key]['parts'] = parts
        self.read_boolean_option(section, "warnings")
        if self.has_option(section, "verbose"):
            if self.getboolean(section, "verbose"):
                self.config["verbose"] = True
                self.config["warnings"] = True
        if self.has_option(section, "quiet"):
            if self.getboolean(section, "quiet"):
                self.config['output'] = 'none'
                self.config['quiet'] = True
                logconf.reset_loglevel()  # if debug will be overwritten next
        if self.has_option(section, "debug"):
            val = self.get(section, "debug")
            parts = [f.strip().lower() for f in val.split(',')]
            logconf.set_debug(parts)
        self.read_boolean_option(section, "status")
        if self.has_option(section, "log"):
            val = self.get(section, "log").strip().lower()
            self.config['output'] = val
        if self.has_option(section, "fileoutput"):
            loggers = self.get(section, "fileoutput").split(",")
            # strip names from whitespace
            loggers = (x.strip().lower() for x in loggers)
            # no file output for the failures and none Logger
            from ..logger import LoggerNames

            loggers = (
                x
                for x in loggers
                if x in LoggerNames and x not in ("failures", "none")
            )
            for val in loggers:
                output = self.config.logger_new(val, fileoutput=1)
                self.config['fileoutput'].append(output)
        if self.has_option(section, "ignoreerrors"):
            for line in read_multiline(self.get(section, "ignoreerrors")):
                parts = line.split(maxsplit=1)
                if len(parts) == 1:
                    parts.append('')
                self.config["ignoreerrors"].append(tuple(
                    re_compile(part) for part in parts
                ))

    def read_checking_config(self):
        """Read configuration options in section "checking"."""
        section = "checking"
        self.read_int_option(section, "threads", min=-1)
        self.config['threads'] = max(0, self.config['threads'])
        self.read_int_option(section, "timeout", min=1)
        self.read_int_option(section, "aborttimeout", min=1)
        self.read_int_option(section, "recursionlevel", min=-1)
        self.read_string_option(section, "nntpserver")
        self.read_string_option(section, "useragent")
        self.read_float_option(section, "maxrequestspersecond", min=0.001)
        self.read_int_option(section, "maxnumurls", min=0)
        self.read_int_option(section, "maxfilesizeparse", min=1)
        self.read_int_option(section, "maxfilesizedownload", min=1)
        if self.has_option(section, "allowedschemes"):
            self.config['allowedschemes'] = [
                x.strip().lower()
                for x in self.get(section, 'allowedschemes').split(',')
            ]
        self.read_boolean_option(section, "debugmemory")
        self.read_string_option(section, "cookiefile")
        self.read_boolean_option(section, "robotstxt")
        self.read_string_option(section, "localwebroot")
        try:
            self.read_boolean_option(section, "sslverify")
        except ValueError:
            self.read_string_option(section, "sslverify")
        self.read_int_option(section, "maxrunseconds", min=0)
        self.read_int_option(section, "resultcachesize", min=0)

    def read_authentication_config(self):
        """Read configuration options in section "authentication"."""
        section = "authentication"
        password_fields = []
        if self.has_option(section, "entry"):
            for val in read_multiline(self.get(section, "entry")):
                auth = val.split()
                if len(auth) == 3:
                    self.config.add_auth(
                        pattern=auth[0], user=auth[1], password=auth[2]
                    )
                    password_fields.append(f"entry/{auth[0]}/{auth[1]}")
                elif len(auth) == 2:
                    self.config.add_auth(pattern=auth[0], user=auth[1])
                else:
                    raise LinkCheckerError(
                        _("missing auth part in entry %(val)r") % {"val": val}
                    )
        # read login URL and field names
        if self.has_option(section, "loginurl"):
            val = self.get(section, "loginurl").strip()
            if not (
                val.lower().startswith("http:") or val.lower().startswith("https:")
            ):
                raise LinkCheckerError(
                    _(
                        "invalid login URL `%s'. Only "
                        "HTTP and HTTPS URLs are supported."
                    )
                    % val
                )
            self.config["loginurl"] = val
        self.read_string_option(section, "loginuserfield")
        self.read_string_option(section, "loginpasswordfield")
        # read login extra fields
        if self.has_option(section, "loginextrafields"):
            for val in read_multiline(self.get(section, "loginextrafields")):
                name, value = val.split(":", 1)
                self.config["loginextrafields"][name] = value
        self.check_password_readable(section, password_fields)

    def check_password_readable(self, section, fields):
        """Check if there is a readable configuration file and print a warning."""
        if not fields:
            return
        # The information which of the  configuration files
        # included which option is not available. To avoid false positives,
        # a warning is only printed if exactly one file has been read.
        if len(self.read_ok) != 1:
            return
        fn = self.read_ok[0]
        if fileutil.is_accessable_by_others(fn):
            log.warn(
                LOG_CHECK,
                _(
                    "The configuration file %s contains password information (in"
                    " section [%s] and options %s) and the file is readable by"
                    " others. Please make the file only readable by you."
                ),
                fn,
                section,
                fields,
            )
            if os.name == 'posix':
                log.warn(LOG_CHECK, _("For example execute 'chmod go-rw %s'.") % fn)
            elif os.name == 'nt':
                log.warn(
                    LOG_CHECK,
                    _(
                        "See %(url)s for more info on setting file permissions."
                    ) % {"url": "https://support.microsoft.com/kb/308419"}
                )

    def read_filtering_config(self):
        """
        Read configuration options in section "filtering".
        """
        section = "filtering"
        if self.has_option(section, "ignorewarnings"):
            self.config['ignorewarnings'] = [
                f.strip().lower()
                for f in self.get(section, 'ignorewarnings').split(',')
            ]
        if self.has_option(section, "ignore"):
            for line in read_multiline(self.get(section, "ignore")):
                pat = get_link_pat(line, strict=1)
                self.config["externlinks"].append(pat)
        if self.has_option(section, "nofollow"):
            for line in read_multiline(self.get(section, "nofollow")):
                pat = get_link_pat(line, strict=0)
                self.config["externlinks"].append(pat)
        if self.has_option(section, "internlinks"):
            pat = get_link_pat(self.get(section, "internlinks"))
            self.config["internlinks"].append(pat)
        self.read_boolean_option(section, "checkextern")

    def read_plugin_config(self):
        """Read plugin-specific configuration values."""
        folders = self.config["pluginfolders"]
        modules = plugins.get_plugin_modules(folders)
        for pluginclass in plugins.get_plugin_classes(modules):
            section = pluginclass.__name__
            if self.has_section(section):
                self.config["enabledplugins"].append(section)
                self.config[section] = pluginclass.read_config(self)



# Copyright (C) 2000-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Handle nntp: and news: links.
"""

import re
import time
import nntplib
import random

from . import urlbase
from .. import log, LinkCheckerError, LOG_CHECK
from .const import WARN_NNTP_NO_SERVER, WARN_NNTP_NO_NEWSGROUP

random.seed()


class NntpUrl(urlbase.UrlBase):
    """
    Url link with NNTP scheme.
    """

    def check_connection(self):
        """
        Connect to NNTP server and try to request the URL article
        resource (if specified).
        """
        nntpserver = self.host or self.aggregate.config["nntpserver"]
        if not nntpserver:
            self.add_warning(
                _("No NNTP server was specified, skipping this URL."),
                tag=WARN_NNTP_NO_SERVER,
            )
            return
        nntp = self._connect_nntp(nntpserver)
        group = self.urlparts[2]
        while group[:1] == '/':
            group = group[1:]
        if '@' in group:
            # request article info (resp, number mid)
            number = nntp.stat("<" + group + ">")[1]
            self.add_info(_('Article number %(num)s found.') % {"num": number})
        else:
            # split off trailing article span
            group = group.split('/', 1)[0]
            if group:
                # request group info (resp, count, first, last, name)
                name = nntp.group(group)[4]
                self.add_info(_("News group %(name)s found.") % {"name": name})
            else:
                # group name is the empty string
                self.add_warning(
                    _("No newsgroup specified in NNTP URL."), tag=WARN_NNTP_NO_NEWSGROUP
                )

    def _connect_nntp(self, nntpserver):
        """
        This is done only once per checking task. Also, the newly
        introduced error codes 504 and 505 (both inclining "Too busy, retry
        later", are caught.
        """
        tries = 0
        nntp = None
        while tries < 2:
            tries += 1
            try:
                nntp = nntplib.NNTP(nntpserver, usenetrc=False)
            except nntplib.NNTPTemporaryError:
                self.wait()
            except nntplib.NNTPPermanentError as msg:
                if re.compile("^50[45]").search(str(msg)):
                    self.wait()
                else:
                    raise
        if nntp is None:
            raise LinkCheckerError(
                _("NNTP server too busy; tried more than %d times.") % tries
            )
        if log.is_debug(LOG_CHECK):
            nntp.set_debuglevel(1)
        self.add_info(nntp.getwelcome())
        return nntp

    def wait(self):
        """Wait some time before trying to connect again."""
        time.sleep(random.randrange(10, 30))

    def can_get_content(self):
        """
        NNTP urls have no content.

        @return: False
        @rtype: bool
        """
        return False



# Copyright (C) 2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Add HTTP server name information
"""
from . import _ConnectionPlugin


class HttpHeaderInfo(_ConnectionPlugin):
    """Add HTTP header info for each URL"""

    def __init__(self, config):
        """Initialize configuration."""
        super().__init__(config)
        self.prefixes = tuple(config["prefixes"])

    def applies_to(self, url_data):
        """Check for HTTP and prefix config."""
        return self.prefixes and url_data.is_http()

    def check(self, url_data):
        """Check content for invalid anchors."""
        headers = []
        for name, value in url_data.headers.items():
            if name.lower().startswith(self.prefixes):
                headers.append(name.lower())
        if headers:
            items = [
                f"{name.capitalize()}={url_data.headers[name]}"
                for name in headers
            ]
            info = "HTTP headers %s" % ", ".join(items)
            url_data.add_info(info)

    @classmethod
    def read_config(cls, configparser):
        """Read configuration file options."""
        config = dict()
        section = cls.__name__
        option = "prefixes"
        if configparser.has_option(section, option):
            value = configparser.get(section, option)
            names = [x.strip().lower() for x in value.split(",")]
        else:
            names = []
        config[option] = names
        return config



# Copyright (C) 2000-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Handle telnet: links.
"""

import telnetlib
from . import urlbase
from .. import log, LOG_CHECK


def encode(s, encoding="iso-8859-1", errors="ignore"):
    """Encode telnet data like username and password."""
    return s.encode(encoding, errors)


class TelnetUrl(urlbase.UrlBase):
    """
    Url link with telnet scheme.
    """

    def build_url(self):
        """
        Call super.build_url(), set default telnet port and initialize
        the login credentials.
        """
        super().build_url()
        # default port
        if self.port is None:
            self.port = 23
        # set user/pass
        self.user, self.password = self.get_user_password()

    def local_check(self):
        """
        Warn about empty host names. Else call super.local_check().
        """
        if not self.host:
            self.set_result(_("Host is empty"), valid=False)
            return
        super().local_check()

    def check_connection(self):
        """
        Open a telnet connection and try to login. Expected login
        label is "login: ", expected password label is "Password: ".
        """
        self.url_connection = telnetlib.Telnet(timeout=self.aggregate.config["timeout"])
        if log.is_debug(LOG_CHECK):
            self.url_connection.set_debuglevel(1)
        self.url_connection.open(self.host, self.port)
        if self.user:
            self.url_connection.read_until(b"login: ", 10)
            self.url_connection.write(encode(self.user) + b"\n")
            if self.password:
                self.url_connection.read_until(b"Password: ", 10)
                self.url_connection.write(encode(self.password) + b"\n")
                # XXX how to tell if we are logged in??
        self.url_connection.write(b"exit\n")

    def can_get_content(self):
        """
        Telnet URLs have no content.

        @return: False
        @rtype: bool
        """
        return False
